import json
from collections.abc import Mapping
from datetime import datetime, timezone
from typing import Any, ClassVar

import pytest
from cognite.client.data_classes import Annotation, Asset, Event, FileMetadata, Sequence, TimeSeries
from cognite.client.data_classes.data_modeling import (
    EdgeApply,
    NodeId,
    NodeOrEdgeData,
)
from cognite.client.data_classes.data_modeling import data_types as dt
from cognite.client.data_classes.data_modeling.data_types import DirectRelationReference, EnumValue
from cognite.client.data_classes.data_modeling.ids import ContainerId, EdgeId, ViewId
from cognite.client.data_classes.data_modeling.instances import PropertyValueWrite
from cognite.client.data_classes.data_modeling.views import MappedProperty, MultiEdgeConnection, ViewProperty

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId, ResourceViewMapping
from cognite_toolkit._cdf_tk.commands._migrate.conversion import (
    DirectRelationCache,
    asset_centric_to_dm,
    create_properties,
)
from cognite_toolkit._cdf_tk.commands._migrate.issues import (
    ConversionIssue,
    FailedConversion,
    InvalidPropertyDataType,
)


class TestCreateProperties:
    INSTANCE_ID = NodeId(space="test_space", external_id="test_instance")
    CONTAINER_ID = ContainerId("test_space", "test_container")
    DEFAULT_CONTAINER_ARGS: ClassVar = dict(nullable=True, immutable=False, auto_increment=False)
    ASSET_CENTRIC_ID = AssetCentricId(resource_type="asset", id_=123)
    EVENT_CENTRIC_ID = AssetCentricId(resource_type="event", id_=456)
    DIRECT_RELATION_CACHE = DirectRelationCache(
        asset={1: DirectRelationReference("instance_space", "MyFirstAsset")},
        source={"sourceA": DirectRelationReference("instance_space", "TheSourceA")},
        file={},
    )

    @pytest.mark.parametrize(
        "dumped,view_properties,property_mapping,expected_properties,expected_issue",
        [
            pytest.param(
                {"name": "MyAsset", "description": "An asset,", "metadata": {"categoryNo": "1"}},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "descriptionId": MappedProperty(CONTAINER_ID, "descriptionId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "categoryNoId": MappedProperty(CONTAINER_ID, "categoryNoId", dt.Int64(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "description": "descriptionId", "metadata.categoryNo": "categoryNoId"},
                {"nameId": "MyAsset", "descriptionId": "An asset,", "categoryNoId": 1},
                ConversionIssue(asset_centric_id=ASSET_CENTRIC_ID, instance_id=INSTANCE_ID),
                id="Basic property mapping with integer conversion and no issues",
            ),
            pytest.param(
                {"name": "MyAsset", "created": "2023-01-01T12:00:00Z", "active": True},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "createdId": MappedProperty(CONTAINER_ID, "createdId", dt.Timestamp(), **DEFAULT_CONTAINER_ARGS),
                    "activeId": MappedProperty(CONTAINER_ID, "activeId", dt.Boolean(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "created": "createdId", "active": "activeId"},
                {
                    "nameId": "MyAsset",
                    "createdId": datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
                    "activeId": True,
                },
                ConversionIssue(asset_centric_id=ASSET_CENTRIC_ID, instance_id=INSTANCE_ID),
                id="Multiple data types conversion",
            ),
            pytest.param(
                {"name": "MyAsset"},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "descriptionId": MappedProperty(CONTAINER_ID, "descriptionId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "description": "descriptionId"},
                {"nameId": "MyAsset"},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    missing_asset_centric_properties=["description"],
                ),
                id="Missing property in flattened dump",
            ),
            pytest.param(
                {"name": "MyAsset", "description": "Test"},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "description": "descriptionId"},
                {"nameId": "MyAsset"},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    missing_instance_properties=["descriptionId"],
                ),
                id="Missing property in view properties",
            ),
            pytest.param(
                {"name": "MyAsset"},
                {
                    "nameId": MultiEdgeConnection(
                        type=DirectRelationReference("space", "Other"),
                        source=ViewId("space", "view"),
                        name=None,
                        description=None,
                        edge_source=None,
                        direction="outwards",
                    ),
                },
                {"name": "nameId"},
                {},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    invalid_instance_property_types=[
                        InvalidPropertyDataType(property_id="nameId", expected_type="MappedProperty")
                    ],
                ),
                id="Invalid property type",
            ),
            pytest.param(
                {"number": "not-a-number"},
                {
                    "numberId": MappedProperty(CONTAINER_ID, "numberId", dt.Int64(), **DEFAULT_CONTAINER_ARGS),
                },
                {"number": "numberId"},
                {},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    failed_conversions=[
                        FailedConversion(
                            property_id="number", value="not-a-number", error="Cannot convert not-a-number to int64."
                        )
                    ],
                ),
                id="Conversion error",
            ),
            pytest.param(
                {},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId"},
                {},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    missing_asset_centric_properties=["name"],
                ),
                id="Empty dictionary",
            ),
            pytest.param(
                {"name": "MyAsset", "labels": [{"externalId": "tag1"}, {"externalId": "tag2"}]},
                {
                    "nameId": MappedProperty(CONTAINER_ID, "nameId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "tags": MappedProperty(
                        ContainerId("cdf_cdm", "CogniteDescribable"),
                        "tags",
                        dt.Text(is_list=True),
                        **DEFAULT_CONTAINER_ARGS,
                    ),
                },
                {"name": "nameId", "labels": "tags"},
                {"nameId": "MyAsset", "tags": ["tag1", "tag2"]},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                ),
                id="List of simple types (labels to list of strings)",
            ),
            pytest.param(
                {
                    "name": "MyAsset",
                    "labels": [{"externalId": "tag1"}, {"externalId": "tag2"}],
                    "metadata": {"key": "value"},
                },
                {
                    "nameId": MappedProperty(CONTAINER_ID, "tags", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "tag": MappedProperty(CONTAINER_ID, "tag", dt.Text(is_list=False), **DEFAULT_CONTAINER_ARGS),
                    "metadata": MappedProperty(CONTAINER_ID, "metadata", dt.Json(), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "labels[0].externalId": "tag", "metadata": "metadata"},
                {"nameId": "MyAsset", "tag": "tag1", "metadata": {"key": "value"}},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["labels[1].externalId"],
                ),
                id="Mapping the first label and entire metadata.",
            ),
            pytest.param(
                {
                    "name": "MyAsset",
                    "type": "TypeA",
                    "metadata": {"category": "TypeB"},
                },
                {
                    "nameId": MappedProperty(CONTAINER_ID, "tags", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "tag": MappedProperty(CONTAINER_ID, "tag", dt.Text(is_list=False), **DEFAULT_CONTAINER_ARGS),
                },
                {"name": "nameId", "type": "tag", "metadata.category": "tag"},
                {"nameId": "MyAsset", "tag": "TypeA"},
                ConversionIssue(
                    asset_centric_id=ASSET_CENTRIC_ID,
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["metadata.category"],
                ),
                id="Duplicated mapping target",
            ),
        ],
    )
    def test_create_properties(
        self,
        dumped: dict[str, Any],
        view_properties: dict[str, ViewProperty],
        property_mapping: dict[str, str],
        expected_properties: dict[str, PropertyValueWrite],
        expected_issue: ConversionIssue,
    ) -> None:
        issue = ConversionIssue(asset_centric_id=self.ASSET_CENTRIC_ID, instance_id=self.INSTANCE_ID)
        properties = create_properties(
            dumped, view_properties, property_mapping, "asset", issue, self.DIRECT_RELATION_CACHE
        )

        assert properties == expected_properties

        assert issue.dump() == expected_issue.dump()

    @pytest.mark.parametrize(
        "dumped,view_properties,property_mapping,expected_properties,expected_issue",
        [
            pytest.param(
                {"startTime": 123, "endTime": 321, "description": "An event", "source": "sourceA", "assetIds": [1]},
                {
                    "startTimeId": MappedProperty(
                        CONTAINER_ID, "startTimeId", dt.Timestamp(), **DEFAULT_CONTAINER_ARGS
                    ),
                    "endTimeId": MappedProperty(CONTAINER_ID, "endTimeId", dt.Timestamp(), **DEFAULT_CONTAINER_ARGS),
                    "descriptionId": MappedProperty(CONTAINER_ID, "descriptionId", dt.Text(), **DEFAULT_CONTAINER_ARGS),
                    "source": MappedProperty(
                        ContainerId("cdf_cdm", "CogniteSourceable"),
                        "source",
                        dt.DirectRelation(),
                        **DEFAULT_CONTAINER_ARGS,
                    ),
                    "assets": MappedProperty(
                        CONTAINER_ID, "assets", dt.DirectRelation(is_list=True), **DEFAULT_CONTAINER_ARGS
                    ),
                },
                {
                    "startTime": "startTimeId",
                    "endTime": "endTimeId",
                    "description": "descriptionId",
                    "source": "source",
                    "assetIds": "assets",
                },
                {
                    "startTimeId": datetime(1970, 1, 1, 0, 0, 0, 123000, tzinfo=timezone.utc),
                    "endTimeId": datetime(1970, 1, 1, 0, 0, 0, 321000, tzinfo=timezone.utc),
                    "descriptionId": "An event",
                    "source": DirectRelationReference("instance_space", "TheSourceA"),
                    "assets": [DirectRelationReference("instance_space", "MyFirstAsset")],
                },
                ConversionIssue(asset_centric_id=EVENT_CENTRIC_ID, instance_id=INSTANCE_ID),
                id="Basic event property mapping with direct relations",
            ),
        ],
    )
    def test_create_properties_events(
        self,
        dumped: dict[str, Any],
        view_properties: dict[str, ViewProperty],
        property_mapping: dict[str, str],
        expected_properties: dict[str, PropertyValueWrite],
        expected_issue: ConversionIssue,
    ) -> None:
        issue = ConversionIssue(asset_centric_id=self.EVENT_CENTRIC_ID, instance_id=self.INSTANCE_ID)
        properties = create_properties(
            dumped, view_properties, property_mapping, "event", issue, self.DIRECT_RELATION_CACHE
        )

        assert properties == expected_properties

        assert issue.dump() == expected_issue.dump()


class TestAssetCentricConversion:
    INSTANCE_ID = NodeId(space="test_space", external_id="test_instance")
    CONTAINER_ID = ContainerId("test_space", "test_container")
    VIEW_ID = ViewId("test_space", "test_view", "v1")
    INSTANCE_SOURCE_VIEW_ID = ViewId("cognite_migration", "InstanceSource", "v1")
    ASSET_INSTANCE_ID_BY_ID: ClassVar[Mapping[int, DirectRelationReference]] = {
        123: DirectRelationReference("test_space", "asset_123_instance")
    }
    SOURCE_SYSTEM_INSTANCE_ID_BY_EXTERNAL_ID: ClassVar[Mapping[str, DirectRelationReference]] = {
        "source_system_1": DirectRelationReference("test_space", "source_system_1_instance")
    }

    @pytest.mark.parametrize(
        "resource,view_source,view_properties,expected_properties,expected_issue",
        [
            pytest.param(
                # Simple Asset with basic mapping
                Asset(id=123, external_id="asset_123", name="Test Asset", description="A test asset"),
                ResourceViewMapping(
                    external_id="asset_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="asset",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={"name": "assetName", "description": "assetDescription"},
                ),
                {
                    "assetName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "assetDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetDescription",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {"assetName": "Test Asset", "assetDescription": "A test asset"},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("asset", 123),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[],
                ),
                id="simple_asset_mapping",
            ),
            pytest.param(
                TimeSeries(
                    id=456,
                    external_id="ts_456",
                    name="Test TimeSeries",
                    description="A test timeseries",
                    unit="celsius",
                    metadata={"sensor_type": "temperature", "location": "room_1"},
                ),
                ResourceViewMapping(
                    external_id="timeseries_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="timeseries",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={
                        "name": "timeseriesName",
                        "unit": "measurementUnit",
                        "metadata.sensor_type": "sensorType",
                        "metadata.location": "deviceLocation",
                    },
                ),
                {
                    "timeseriesName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "timeseriesName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "measurementUnit": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "measurementUnit",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "sensorType": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "sensorType",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "deviceLocation": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "deviceLocation",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {
                    "timeseriesName": "Test TimeSeries",
                    "measurementUnit": "celsius",
                    "sensorType": "temperature",
                    "deviceLocation": "room_1",
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("timeseries", id_=456),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                ),
                id="timeseries_with_metadata",
            ),
            pytest.param(
                Event(
                    id=789,
                    external_id="event_789",
                    start_time=1756359489386,
                    end_time=1756359499880,
                    description="Not mapped",
                    source="source_system_1",
                    asset_ids=[123],
                    metadata={
                        "operator": "John Doe",
                        "severity": "HIGH",
                        "value": "invalid_int",  # This will cause a conversion issue
                        "aConnectionProp": json.dumps(
                            {"externalId": "op_123", "space": "schema_space", "type": "Operation"}
                        ),
                    },
                ),
                ResourceViewMapping(
                    external_id="incomplete_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="event",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={
                        "source": "source",
                        "assetIds": "assets",
                        "missing_prop": "targetProp",
                        "startTime": "eventStart",
                        "endTime": "eventEnd",
                        "metadata.operator": "missingDMProp",
                        "metadata.severity": "eventSeverity",
                        "metadata.value": "eventValue",
                        "metadata.missingMetaProp": "anotherMissingDMProp",
                        "metadata.aConnectionProp": "some_other_event",
                    },
                ),
                {
                    "source": MappedProperty(
                        ContainerId("cdf_cdm", "CogniteSourceable"),
                        "source",
                        dt.DirectRelation(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "assets": MappedProperty(
                        ContainerId("cdf_cdm", "CogniteFile"),
                        "assets",
                        dt.DirectRelation(is_list=True, max_list_size=1200),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventStart": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventStart",
                        dt.Timestamp(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventEnd": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventEnd",
                        dt.Timestamp(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventSeverity": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventSeverity",
                        dt.Enum({"high": EnumValue(), "low": EnumValue(), "medium": EnumValue()}),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventValue": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventValue",
                        dt.Int64(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "some_other_event": MultiEdgeConnection(
                        type=DirectRelationReference("schema_space", "Operation"),
                        source=ViewId("test_space", "test_view", "v1"),
                        name=None,
                        description=None,
                        edge_source=None,
                        direction="outwards",
                    ),
                },
                {
                    "eventStart": datetime(2025, 8, 28, 5, 38, 9, 386000, tzinfo=timezone.utc),
                    "eventEnd": datetime(2025, 8, 28, 5, 38, 19, 880000, tzinfo=timezone.utc),
                    "eventSeverity": "high",
                    "source": DirectRelationReference("test_space", "source_system_1_instance"),
                    "assets": [DirectRelationReference("test_space", "asset_123_instance")],
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("event", id_=789),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                    missing_asset_centric_properties=["metadata.missingMetaProp", "missing_prop"],
                    missing_instance_properties=["anotherMissingDMProp", "missingDMProp", "targetProp"],
                    invalid_instance_property_types=[
                        InvalidPropertyDataType(property_id="some_other_event", expected_type="MappedProperty")
                    ],
                    failed_conversions=[
                        FailedConversion(
                            property_id="metadata.value",
                            value="invalid_int",
                            error="Cannot convert invalid_int to int64.",
                        )
                    ],
                ),
                id="Event with conversion issues (missing properties)",
            ),
            pytest.param(
                FileMetadata(
                    id=321,
                    external_id="file_321",
                    name="Test File",
                    mime_type="application/octet-stream",
                    metadata={"file_type": "pdf", "confidential": "true"},
                ),
                ResourceViewMapping(
                    external_id="file_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="file",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={},
                ),
                {
                    "fileName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "fileDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileDescription",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "fileType": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileType",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "isConfidential": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "isConfidential",
                        dt.Boolean(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("file", id_=321),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[
                        "metadata.confidential",
                        "metadata.file_type",
                        "mimeType",
                        "name",
                    ],
                    missing_asset_centric_properties=[],
                    missing_instance_properties=[],
                ),
                id="FileMetadata with no mappings (all ignored)",
            ),
            pytest.param(
                TimeSeries(
                    id=654,
                    name="Test TimeSeries",
                    description="A test timeseries",
                    metadata=None,
                ),
                ResourceViewMapping(
                    external_id="timeseries_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="timeseries",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={"name": "timeSeriesName", "metadata.category": "timeSeriesCategory"},
                ),
                {
                    "timeSeriesName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "timeSeriesName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "timeSeriesCategory": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "timeSeriesCategory",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {"timeSeriesName": "Test TimeSeries"},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("timeseries", id_=654),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                    missing_asset_centric_properties=["metadata.category"],
                    missing_instance_properties=[],
                ),
                id="TimeSeries with partial mapping",
            ),
            pytest.param(
                Asset(id=999, external_id="asset_999", name=None, description=None),
                ResourceViewMapping(
                    external_id="empty_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="asset",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={"name": "assetName", "description": "assetDescription"},
                ),
                {
                    "assetName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetName",
                        dt.Text(),
                        nullable=False,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "assetDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetDescription",
                        dt.Text(),
                        nullable=False,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("asset", id_=999),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[],
                    # Name and description set to None is the same as missing as we have now way of knowing
                    # whether they were explicitly set to None or just not set at all.
                    missing_asset_centric_properties=["description", "name"],
                    missing_instance_properties=[],
                ),
                id="Asset with non-nullable properties all None",
            ),
            pytest.param(
                Event(
                    id=999,
                    external_id="event_999",
                    type="MyType",
                    metadata={"category": "MyCategory"},
                    source="not_existing",
                ),
                ResourceViewMapping(
                    external_id="event_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="event",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    property_mapping={"type": "category", "metadata.category": "category", "source": "source"},
                ),
                {
                    "category": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "category",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "source": MappedProperty(
                        ContainerId("cdf_cdm", "CogniteSourceable"),
                        "source",
                        dt.DirectRelation(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {
                    "category": "MyType",
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("event", id_=999),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["metadata.category"],
                    failed_conversions=[
                        FailedConversion(
                            property_id="source",
                            value="not_existing",
                            error="Cannot convert 'not_existing' to "
                            "DirectRelationReference. Invalid data type "
                            "or missing in lookup.",
                        )
                    ],
                ),
                id="Event with overlapping property and metadata mapping (property takes precedence) and missing source",
            ),
        ],
    )
    def test_asset_centric_to_dm(
        self,
        resource: Asset | FileMetadata | Event | TimeSeries | Sequence,
        view_source: ResourceViewMapping,
        view_properties: dict[str, ViewProperty],
        expected_properties: dict[str, str],
        expected_issue: ConversionIssue,
    ) -> None:
        actual, issue = asset_centric_to_dm(
            resource,
            self.INSTANCE_ID,
            view_source,
            view_properties,
            self.ASSET_INSTANCE_ID_BY_ID,
            self.SOURCE_SYSTEM_INSTANCE_ID_BY_EXTERNAL_ID,
            {},
        )

        # Check the structure of the returned NodeApply
        assert actual.space == self.INSTANCE_ID.space
        assert actual.external_id == self.INSTANCE_ID.external_id
        assert 1 <= len(actual.sources) <= 2

        # Check the main view source
        if len(actual.sources) == 2:
            main_source = actual.sources[0]
            assert main_source.source == view_source.view_id
            assert main_source.properties == expected_properties
        else:
            assert expected_properties == {}

        # Check the instance source view
        instance_source = actual.sources[-1]
        assert instance_source.source == self.INSTANCE_SOURCE_VIEW_ID
        assert instance_source.properties["resourceType"] == view_source.resource_type
        assert instance_source.properties["id"] == resource.id
        if hasattr(resource, "data_set_id") and resource.data_set_id is not None:
            assert instance_source.properties["dataSetId"] == resource.data_set_id
        if hasattr(resource, "external_id") and resource.external_id is not None:
            assert instance_source.properties["classicExternalId"] == resource.external_id

        assert expected_issue.dump() == issue.dump()

    ANNOTATION_ID = EdgeId(space="test_space", external_id="annotation_37")
    DEFAULT_PROPERTIES: ClassVar[dict[str, Any]] = dict(nullable=True, immutable=False, auto_increment=False)

    DIAGRAM_ANNOTATION_PROPERTIES: ClassVar[dict[str, ViewProperty]] = {
        "status": MappedProperty(
            container=CONTAINER_ID,
            container_property_identifier="status",
            type=dt.Enum({"Suggested": EnumValue(), "Approved": EnumValue(), "Rejected": EnumValue()}),
            **DEFAULT_PROPERTIES,
        ),
    }
    ANNOTATION_MAPPING = ResourceViewMapping(
        external_id="file_annotation_mapping",
        view_id=ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
        property_mapping={
            "annotatedResourceId": "edge.startNode",
            "annotationType": "edge.type.externalId",
            "status": "status",
            "data.assetRef.id": "edge.endNode",
            "data.assetRef.externalId": "edge.endNode",
            "data.text": "edge.invalidProp",
        },
        version=1,
        last_updated_time=1000000,
        created_time=1000000,
        resource_type="fileAnnotation",
    )
    FILE_INSTANCE_BY_ID: ClassVar[Mapping[int, DirectRelationReference]] = {
        42: DirectRelationReference("test_space", "file_456_instance")
    }

    @pytest.mark.parametrize(
        "resource,mapping,expected_edge,expected_issue",
        [
            pytest.param(
                Annotation(
                    id=37,
                    annotated_resource_type="file",
                    annotation_type="diagrams.FileLink",
                    annotated_resource_id=42,
                    creating_user="user_1",
                    creating_app="app_1",
                    creating_app_version="1.0.0",
                    status="Approved",
                    data=dict(assetRef=dict(id=123)),
                ),
                ANNOTATION_MAPPING,
                EdgeApply(
                    space=ANNOTATION_ID.space,
                    external_id=ANNOTATION_ID.external_id,
                    start_node=DirectRelationReference("test_space", "file_456_instance"),
                    end_node=DirectRelationReference("test_space", "asset_123_instance"),
                    type=DirectRelationReference("test_space", "diagrams.FileLink"),
                    sources=[
                        NodeOrEdgeData(
                            source=ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
                            properties={"status": "Approved"},
                        )
                    ],
                ),
                ConversionIssue(
                    asset_centric_id=AssetCentricId("fileAnnotation", id_=37),
                    instance_id=EdgeId(space="test_space", external_id="annotation_37"),
                    ignored_asset_centric_properties=[
                        "annotatedResourceType",
                        "creatingApp",
                        "creatingAppVersion",
                        "creatingUser",
                    ],
                    missing_asset_centric_properties=["data.assetRef.externalId", "data.text"],
                ),
                id="Basic annotation conversion",
            )
        ],
    )
    def test_asset_centric_to_annotation(
        self,
        resource: Annotation,
        mapping: ResourceViewMapping,
        expected_edge: EdgeApply,
        expected_issue: ConversionIssue,
    ) -> None:
        """Testing that asset_centric_to_dm raises can convert asset annotations. Note that unlike the other resource types,
        we do not track the linage of annotations."""

        edge, issue = asset_centric_to_dm(
            resource,
            self.ANNOTATION_ID,
            mapping,
            self.DIAGRAM_ANNOTATION_PROPERTIES,
            self.ASSET_INSTANCE_ID_BY_ID,
            {},
            self.FILE_INSTANCE_BY_ID,
        )

        assert edge.dump() == expected_edge.dump()
        assert issue.dump() == expected_issue.dump()

    def test_asset_centric_to_annotation_failed(self) -> None:
        """Testing that asset_centric_to_dm raises conversion issues for annotations with missing required properties."""

        resource = Annotation(
            id=38,
            annotated_resource_type="file",
            annotation_type="diagrams.FileLink",
            annotated_resource_id=999,  # Missing file instance
            creating_user="user_1",
            creating_app="app_1",
            creating_app_version="1.0.0",
            status="Approved",
            data=dict(assetRef=dict(id=123), text="Some annotation text"),
        )

        edge, issue = asset_centric_to_dm(
            resource,
            EdgeId(space="test_space", external_id="annotation_38"),
            self.ANNOTATION_MAPPING,
            self.DIAGRAM_ANNOTATION_PROPERTIES,
            self.ASSET_INSTANCE_ID_BY_ID,
            {},
            self.FILE_INSTANCE_BY_ID,
        )

        expected_issue = ConversionIssue(
            asset_centric_id=AssetCentricId("fileAnnotation", id_=38),
            instance_id=EdgeId(space="test_space", external_id="annotation_38"),
            ignored_asset_centric_properties=[
                "annotatedResourceType",
                "creatingApp",
                "creatingAppVersion",
                "creatingUser",
            ],
            missing_asset_centric_properties=["data.assetRef.externalId"],
            missing_instance_properties=[],
            invalid_instance_property_types=[
                InvalidPropertyDataType(property_id="edge.invalidProp", expected_type="EdgeProperty")
            ],
            failed_conversions=[
                FailedConversion(
                    property_id="annotatedResourceId",
                    value=999,
                    error="Cannot convert 999 to DirectRelationReference. Invalid data type or missing in lookup.",
                )
            ],
        )

        assert issue.dump() == expected_issue.dump()
        assert edge is None



import random
import time
from collections.abc import Iterable
from pathlib import Path

import pytest
from cognite.client.data_classes import FileMetadataList, FileMetadataWrite
from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import FILE_METADATA_ID
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import AssetCentricMigrationIO
from cognite_toolkit._cdf_tk.commands._migrate.selectors import MigrationCSVFileSelector


@pytest.fixture()
def three_files_with_content(
    toolkit_client_with_pending_ids: ToolkitClient, toolkit_space: Space
) -> Iterable[FileMetadataList]:
    client = toolkit_client_with_pending_ids
    space = toolkit_space.space
    files = FileMetadataList([])
    for i in range(3):
        file = FileMetadataWrite(
            external_id=f"toolkit_test_migration_{i}_{random.randint(0, 10_000)!s}",
            name=f"toolkit_test_migration_{i}",
        )
        files.append(file)
    output = client.files.retrieve_multiple(external_ids=files.as_external_ids(), ignore_unknown_ids=True)
    if output:
        try:
            client.files.delete(external_id=output.as_external_ids(), ignore_unknown_ids=True)
        except CogniteAPIError:
            client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in output])
    created = FileMetadataList([])
    for file in files:
        uploaded = client.files.upload_bytes(b"test content", **file.dump(camel_case=False))
        created.append(uploaded)

    yield created

    # Cleanup after test
    deleted = client.data_modeling.instances.delete([NodeId(space, file.external_id) for file in created])
    if deleted.nodes:
        return
    client.files.delete(external_id=created.as_external_ids(), ignore_unknown_ids=True)


class TestMigrateFilesCommand:
    def test_migrate_files_v2(
        self,
        toolkit_client_with_pending_ids: ToolkitClient,
        three_files_with_content: FileMetadataList,
        toolkit_space: Space,
        tmp_path: Path,
    ):
        client = toolkit_client_with_pending_ids
        space = toolkit_space.space

        input_file = tmp_path / "files_migration.csv"
        with input_file.open("w", encoding="utf-8") as f:
            f.write(
                "id,space,externalId,ingestionView\n"
                + "\n".join(f"{ts.id},{space},{ts.external_id},{FILE_METADATA_ID}" for ts in three_files_with_content)
                + "\n"
            )

        cmd = MigrationCommand(skip_tracking=True, silent=True)
        results = cmd.migrate(
            selected=MigrationCSVFileSelector(datafile=input_file, kind="FileMetadata"),
            data=AssetCentricMigrationIO(client, skip_linking=False),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=False,
        )
        actual_results = [results.get_progress(f"file_{item.id}") for item in three_files_with_content]
        expected_results = [
            {
                cmd.Steps.DOWNLOAD: "success",
                cmd.Steps.CONVERT: "success",
                cmd.Steps.UPLOAD: "success",
            }
            for _ in three_files_with_content
        ]
        assert actual_results == expected_results

        # Wait for syncer by polling
        for _ in range(12):  # Poll for up to 60 seconds
            migrated_files = client.files.retrieve_multiple(external_ids=three_files_with_content.as_external_ids())
            if all(f.instance_id is not None for f in migrated_files):
                break
            time.sleep(5)

        missing_node_id = [ts.external_id for ts in migrated_files if ts.instance_id is None]
        assert not missing_node_id, f"Some files are missing NodeId: {missing_node_id}"

        node_ids = [ts.instance_id for ts in migrated_files]
        for node_id in node_ids:
            content = client.files.download_bytes(instance_id=node_id)
            assert content == b"test content", f"Content of file {node_id} does not match expected content."



from __future__ import annotations

from typing import Any

import pytest
import yaml

from cognite_toolkit._cdf_tk.data_classes import (
    ConfigEntry,
    Environment,
    InitConfigYAML,
)
from cognite_toolkit._cdf_tk.utils import YAMLComment, flatten_dict
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump
from tests.data import PROJECT_FOR_TEST


class TestConfigYAML:
    def test_producing_correct_keys(
        self, project_for_test_config_dev_yaml: str, dummy_environment: Environment
    ) -> None:
        expected_keys = set(flatten_dict(yaml.safe_load(project_for_test_config_dev_yaml)))
        # Custom keys are not loaded from the module folder.
        # This custom key is added to the dev.config.yaml for other tests.
        expected_keys.remove(("variables", "custom_modules", "my_example_module", "transformation_is_paused"))
        # Skip all environment variables
        expected_keys = {k for k in expected_keys if not k[0] == "environment"}

        config = InitConfigYAML(dummy_environment).load_defaults(PROJECT_FOR_TEST)

        actual_keys = set(config.keys())
        missing = expected_keys - actual_keys
        assert not missing, f"Missing keys: {missing}"
        extra = actual_keys - expected_keys
        assert not extra, f"Extra keys: {extra}"

    def test_extract_extract_config_yaml_comments(self, project_for_test_config_dev_yaml: str) -> None:
        expected_comments = {
            ("variables", "modules", "a_module", "readonly_source_id"): YAMLComment(
                above=["This is a comment in the middle of the file"], after=[]
            ),
            ("variables", "modules", "another_module", "source_asset"): YAMLComment(
                above=["This is a comment at the beginning of the module."],
                after=["This is an extra comment added to the config only 'lore ipsum'"],
            ),
            ("variables", "modules", "another_module", "source_files"): YAMLComment(
                after=["This is a comment after a variable"]
            ),
        }

        actual_comments = InitConfigYAML._extract_comments(project_for_test_config_dev_yaml)

        assert actual_comments == expected_comments

    @pytest.mark.parametrize(
        "raw_file, key_prefix, expected_comments",
        [
            pytest.param(
                """---
# This is a module comment
variable: value # After variable comment
# Before variable comment
variable2: value2
variable3: 'value with #in it'
variable4: "value with #in it" # But a comment after
""",
                tuple("super_module.module_a".split(".")),
                {
                    ("super_module", "module_a", "variable"): YAMLComment(
                        after=["After variable comment"], above=["This is a module comment"]
                    ),
                    ("super_module", "module_a", "variable2"): YAMLComment(above=["Before variable comment"]),
                    ("super_module", "module_a", "variable4"): YAMLComment(after=["But a comment after"]),
                },
                id="module comments",
            )
        ],
    )
    def test_extract_default_config_comments(
        self, raw_file: str, key_prefix: tuple[str, ...], expected_comments: dict[str, Any]
    ):
        actual_comments = InitConfigYAML._extract_comments(raw_file, key_prefix)
        assert actual_comments == expected_comments

    def test_persist_variable_with_comment(self, project_for_test_config_dev_yaml: str) -> None:
        custom_comment = "This is an extra comment added to the config only 'lore ipsum'"

        config = InitConfigYAML.load_existing(project_for_test_config_dev_yaml).load_defaults(PROJECT_FOR_TEST)

        dumped = config.dump_yaml_with_comments()
        loaded = yaml.safe_load(dumped)
        assert loaded["variables"]["modules"]["another_module"]["source_asset"] == "my_new_workmate"
        assert custom_comment in dumped

    def test_added_and_removed_variables(self, project_for_test_config_dev_yaml: str) -> None:
        existing_config_yaml = yaml.safe_load(project_for_test_config_dev_yaml)
        # Added = Exists in the BUILD_CONFIG directory default.config.yaml files but not in config.yaml
        existing_config_yaml["variables"]["modules"]["another_module"].pop("source_asset")
        # Removed = Exists in config.yaml but not in the BUILD_CONFIG directory default.config.yaml files
        existing_config_yaml["variables"]["modules"]["another_module"]["removed_variable"] = "old_value"

        config = InitConfigYAML.load_existing(yaml_safe_dump(existing_config_yaml)).load_defaults(PROJECT_FOR_TEST)

        removed = [v for v in config.values() if v.default_value is None]
        # There is already a custom variable in the config.yaml file
        assert len(removed) == 2
        assert ("variables", "modules", "another_module", "removed_variable") in [v.key_path for v in removed]

        added = [v for v in config.values() if v.current_value is None]
        assert len(added) == 1
        assert added[0].key_path == ("variables", "modules", "another_module", "source_asset")

    def test_load_variables(self, dummy_environment: Environment) -> None:
        expected = {
            ("variables", "modules", "a_module", "readonly_source_id"),
            # default_location is used in two modules and is moved to the top level
            ("variables", "modules", "default_location"),
            ("variables", "modules", "another_module", "source_files"),
            ("variables", "modules", "another_module", "model_space"),
            ("variables", "modules", "parent_module", "child_module", "source_asset"),
        }

        config = InitConfigYAML(dummy_environment).load_variables(PROJECT_FOR_TEST, propagate_reused_variables=True)

        missing = expected - set(config.keys())
        extra = set(config.keys()) - expected
        assert not missing, f"Missing keys: {missing}. Got extra {extra}"
        assert not extra, f"Extra keys: {extra}"

    def test_load_parent_variables(self, dummy_environment: Environment) -> None:
        config = InitConfigYAML(
            dummy_environment,
            {
                ("variables", "modules", "infield", "shared_variable"): ConfigEntry(
                    key_path=("variables", "modules", "infield", "shared_variable"),
                    default_value="shared_value",
                )
            },
        )

        config._load_variables({"shared_variable": {("cognite_modules", "infield", "cdf_infield_common")}})

        assert ("variables", "modules", "infield", "shared_variable") in config.keys()
        assert ("variables", "modules", "infield", "cdf_infield_common", "shared_variable") not in config.keys()

    def test_trailing_slash_in_selected(self, project_for_test_config_dev_yaml: str) -> None:
        existing_config = InitConfigYAML.load_existing(project_for_test_config_dev_yaml).load_defaults(PROJECT_FOR_TEST)
        dumped_config = existing_config.dump_yaml_with_comments()

        dumped_config_dict = yaml.safe_load(dumped_config)

        assert dumped_config_dict["environment"]["selected"][0] == "modules/"

    def test_ignoring_patterns(self, dummy_environment: Environment) -> None:
        ignore_pattern = [
            ("modules", "parent_module", "*", "child_variable"),
            ("modules", "a_module", "readwrite_source_id"),
        ]

        config = InitConfigYAML(dummy_environment).load_defaults(PROJECT_FOR_TEST)
        config_with_ignore = InitConfigYAML(dummy_environment).load_defaults(
            PROJECT_FOR_TEST, ignore_patterns=ignore_pattern
        )

        ignored = set(config.keys()) - set(config_with_ignore.keys())
        assert ignored == {
            ("variables", "modules", "parent_module", "child_module", "child_variable"),
            ("variables", "modules", "a_module", "readwrite_source_id"),
        }

    def test_lift_variables(self, dummy_environment) -> None:
        config = InitConfigYAML(dummy_environment).load_defaults(PROJECT_FOR_TEST)

        config.lift()

        assert ("variables", "modules", "dataset") in config.keys()
        assert ("variables", "modules", "a_module", "dataset") not in config.keys()
        assert ("variables", "modules", "another_module", "dataset") not in config.keys()
        assert ("variables", "modules", "another_module", "source_asset") in config.keys()
        assert ("variables", "modules", "parent_module", "child_module", "source_asset") in config.keys()

    def test_allow_wide_lines(self, dummy_environment) -> None:
        long_default_value = "value " * 20  # whitespace makes default yaml.dump split lines

        config = InitConfigYAML(
            dummy_environment,
            {
                ("variables", "modules", "infield", "shared_variable"): ConfigEntry(
                    key_path=("variables", "modules", "infield", "shared_variable"),
                    default_value=long_default_value,
                )
            },
        )

        dumped = config.dump_yaml_with_comments()
        loaded = yaml.safe_load(dumped)
        assert loaded["variables"]["modules"]["infield"]["shared_variable"] == long_default_value



import gzip
import random
import sys
import time
from collections import deque
from collections.abc import MutableMapping, Sequence, Set
from typing import Literal

import httpx
from cognite.client import global_config
from rich.console import Console

from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils.auxiliary import get_current_toolkit_version, get_user_agent
from cognite_toolkit._cdf_tk.utils.http_client._data_classes import (
    BodyRequest,
    DataBodyRequest,
    FailedRequestMessage,
    HTTPMessage,
    ItemsRequest,
    ParamRequest,
    RequestMessage,
    ResponseList,
    ResponseMessage,
)
from cognite_toolkit._cdf_tk.utils.useful_types import PrimitiveType

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from cognite_toolkit._cdf_tk.client.config import ToolkitClientConfig


class HTTPClient:
    """An HTTP client.

    This class handles rate limiting, retries, and error handling for HTTP requests.

    Args:
        config (ToolkitClientConfig): Configuration for the Toolkit client.
        pool_connections (int): The number of connection pools to cache. Default is 10.
        pool_maxsize (int): The maximum number of connections to save in the pool. Default
            is 20.
        max_retries (int): The maximum number of retries for a request. Default is 10.
        retry_status_codes (frozenset[int]): HTTP status codes that should trigger a retry.
            Default is {408, 429, 502, 503, 504}.
        split_items_status_codes (frozenset[int]): In the case of ItemRequest with multiple
            items, these status codes will trigger splitting the request into smaller batches.

    """

    def __init__(
        self,
        config: ToolkitClientConfig,
        max_retries: int = 10,
        pool_connections: int = 10,
        pool_maxsize: int = 20,
        retry_status_codes: Set[int] = frozenset({408, 429, 502, 503, 504}),
        split_items_status_codes: Set[int] = frozenset({400, 408, 409, 422, 502, 503, 504}),
    ):
        self.config = config
        self._max_retries = max_retries
        self._pool_connections = pool_connections
        self._pool_maxsize = pool_maxsize
        self._retry_status_codes = retry_status_codes
        self._split_items_status_codes = split_items_status_codes

        # Thread-safe session for connection pooling
        self.session = self._create_thread_safe_session()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: object | None
    ) -> Literal[False]:
        """Close the session when exiting the context."""
        self.session.close()
        return False  # Do not suppress exceptions

    def request(self, message: RequestMessage, console: Console | None = None) -> Sequence[HTTPMessage]:
        """Send an HTTP request and return the response.

        Args:
            message (RequestMessage): The request message to send.
            console (Console | None): The rich console to use for printing warnings.

        Returns:
            Sequence[HTTPMessage]: The response message(s). This can also
                include RequestMessage(s) to be retried.
        """
        if isinstance(message, ItemsRequest) and message.tracker and message.tracker.limit_reached():
            error_msg = (
                f"Aborting further splitting of requests after {message.tracker.failed_split_count} failed attempts."
            )
            return message.create_failed_request(error_msg)
        try:
            response = self._make_request(message)
            results = self._handle_response(response, message, console)
        except Exception as e:
            results = self._handle_error(e, message)
        return results

    def request_with_retries(self, message: RequestMessage, console: Console | None = None) -> ResponseList:
        """Send an HTTP request and handle retries.

        This method will keep retrying the request until it either succeeds or
        exhausts the maximum number of retries.

        Note this method will use the current thread to process all request, thus
        it is blocking.

        Args:
            message (RequestMessage): The request message to send.
            console (Console | None): The rich console to use for printing warnings.

        Returns:
            Sequence[ResponseMessage | FailedRequestMessage]: The final response
                messages, which can be either successful responses or failed requests.
        """
        if message.total_attempts > 0:
            raise RuntimeError(f"RequestMessage has already been attempted {message.total_attempts} times.")
        pending_requests: deque[RequestMessage] = deque()
        pending_requests.append(message)
        final_responses = ResponseList([])
        while pending_requests:
            current_request = pending_requests.popleft()
            results = self.request(current_request, console)

            for result in results:
                if isinstance(result, RequestMessage):
                    pending_requests.append(result)
                elif isinstance(result, ResponseMessage | FailedRequestMessage):
                    final_responses.append(result)
                else:
                    raise TypeError(f"Unexpected result type: {type(result)}")

        return final_responses

    def _create_thread_safe_session(self) -> httpx.Client:
        return httpx.Client(
            limits=httpx.Limits(
                max_connections=self._pool_maxsize,
                max_keepalive_connections=self._pool_connections,
            ),
            timeout=self.config.timeout,
        )

    def _create_headers(
        self, api_version: str | None = None, content_type: str = "application/json", accept: str = "application/json"
    ) -> MutableMapping[str, str]:
        headers: MutableMapping[str, str] = {}
        headers["User-Agent"] = f"httpx/{httpx.__version__} {get_user_agent()}"
        auth_name, auth_value = self.config.credentials.authorization_header()
        headers[auth_name] = auth_value
        headers["content-type"] = content_type
        headers["accept"] = accept
        headers["x-cdp-sdk"] = f"CogniteToolkit:{get_current_toolkit_version()}"
        headers["x-cdp-app"] = self.config.client_name
        headers["cdf-version"] = api_version or self.config.api_subversion
        if not global_config.disable_gzip:
            headers["Content-Encoding"] = "gzip"
        return headers

    def _make_request(self, item: RequestMessage) -> httpx.Response:
        headers = self._create_headers(item.api_version, item.content_type, item.accept)
        params: dict[str, PrimitiveType] | None = None
        if isinstance(item, ParamRequest):
            params = item.parameters
        data: str | bytes | None = None
        if isinstance(item, BodyRequest):
            data = item.data()
            if not global_config.disable_gzip:
                data = gzip.compress(data.encode("utf-8"))
        elif isinstance(item, DataBodyRequest):
            data = item.data()
            if not global_config.disable_gzip:
                data = gzip.compress(data)
        return self.session.request(
            method=item.method,
            url=item.endpoint_url,
            content=data,
            headers=headers,
            params=params,
            timeout=self.config.timeout,
            follow_redirects=False,
        )

    def _handle_response(
        self, response: httpx.Response, request: RequestMessage, console: Console | None = None
    ) -> Sequence[HTTPMessage]:
        if 200 <= response.status_code < 300:
            return request.create_success_response(response)
        elif (
            isinstance(request, ItemsRequest)
            and len(request.items) > 1
            and response.status_code in self._split_items_status_codes
        ):
            # 4XX: Status there is at least one item that is invalid, split the batch to get all valid items processed
            # 5xx: Server error, split to reduce the number of items in each request, and count as a status attempt
            status_attempts = request.status_attempt
            if 500 <= response.status_code < 600:
                status_attempts += 1
            splits = request.split(status_attempts=status_attempts)
            if splits[0].tracker and splits[0].tracker.limit_reached():
                return request.create_failure_response(response)
            return splits

        retry_after = self._get_retry_after_in_header(response)
        if retry_after is not None and response.status_code == 429 and request.status_attempt < self._max_retries:
            if console is not None:
                short_url = request.endpoint_url.removeprefix(self.config.base_api_url)
                HighSeverityWarning(
                    f"Rate limit exceeded for the {short_url!r} endpoint. Retrying after {retry_after} seconds."
                ).print_warning(console=console)
            request.status_attempt += 1
            time.sleep(retry_after)
            return [request]

        if request.status_attempt < self._max_retries and response.status_code in self._retry_status_codes:
            request.status_attempt += 1
            time.sleep(self._backoff_time(request.total_attempts))
            return [request]
        else:
            # Permanent failure
            return request.create_failure_response(response)

    @staticmethod
    def _get_retry_after_in_header(response: httpx.Response) -> float | None:
        if "Retry-After" not in response.headers:
            return None
        try:
            return float(response.headers["Retry-After"])
        except ValueError:
            # Ignore invalid Retry-After header
            return None

    @staticmethod
    def _backoff_time(attempts: int) -> float:
        backoff_time = 0.5 * (2**attempts)
        return min(backoff_time, global_config.max_retry_backoff) * random.uniform(0, 1.0)

    def _handle_error(
        self,
        e: Exception,
        request: RequestMessage,
    ) -> Sequence[HTTPMessage]:
        if isinstance(e, httpx.ReadTimeout | httpx.TimeoutException):
            error_type = "read"
            request.read_attempt += 1
            attempts = request.read_attempt
        elif isinstance(e, ConnectionError | httpx.ConnectError | httpx.ConnectTimeout):
            error_type = "connect"
            request.connect_attempt += 1
            attempts = request.connect_attempt
        else:
            error_msg = f"Unexpected exception: {e!s}"
            return request.create_failed_request(error_msg)

        if attempts <= self._max_retries:
            time.sleep(self._backoff_time(request.total_attempts))
            return [request]
        else:
            # We have already incremented the attempt count, so we subtract 1 here
            error_msg = f"RequestException after {request.total_attempts - 1} attempts ({error_type} error): {e!s}"

            return request.create_failed_request(error_msg)



from collections.abc import Iterable, Mapping, Sequence
from types import MappingProxyType
from typing import ClassVar

from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    ContainerList,
    EdgeApply,
    NodeApply,
    SpaceList,
    ViewId,
    ViewList,
)
from cognite.client.data_classes.data_modeling.instances import Instance, InstanceApply
from cognite.client.utils._identifier import InstanceId

from cognite_toolkit._cdf_tk import constants
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceList
from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, SpaceCRUD, ViewCRUD
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.cdf import iterate_instances
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from . import StorageIOConfig
from ._base import ConfigurableStorageIO, Page, UploadableStorageIO
from .selectors import InstanceFileSelector, InstanceSelector, InstanceSpaceSelector, InstanceViewSelector


class InstanceIO(
    ConfigurableStorageIO[InstanceSelector, Instance],
    UploadableStorageIO[InstanceSelector, Instance, InstanceApply],
):
    """This class provides functionality to interact with instances in Cognite Data Fusion (CDF).

    It is used to download, upload, and purge instances, as well as spaces,views, and containers related to instances.

    Args:
        client (ToolkitClient): An instance of ToolkitClient to interact with the CDF API.
        remove_existing_version (bool): Whether to remove existing versions of instances during upload.
            Default is True. Existing versions are used to safeguard against accidental overwrites.

    """

    KIND = "Instances"
    DISPLAY_NAME = "Instances"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = "/models/instances"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = MappingProxyType({"autoCreateDirectRelations": True})
    BASE_SELECTOR = InstanceSelector

    def __init__(self, client: ToolkitClient, remove_existing_version: bool = True) -> None:
        super().__init__(client)
        self._remove_existing_version = remove_existing_version
        # Cache for view to read-only properties mapping
        self._view_readonly_properties_cache: dict[ViewId, set[str]] = {}
        self._view_crud = ViewCRUD.create_loader(self.client)

    def as_id(self, item: Instance) -> str:
        return f"{item.space}:{item.external_id}"

    def _filter_readonly_properties(self, instance: InstanceApply) -> None:
        """
        Filter out read-only properties from the instance.

        Args:
            instance: The instance to filter readonly properties from
        """

        for source in instance.sources:
            readonly_properties = set()
            if isinstance(source.source, ViewId):
                if source.source not in self._view_readonly_properties_cache:
                    self._view_readonly_properties_cache[source.source] = self._view_crud.get_readonly_properties(
                        source.source
                    )
                readonly_properties = self._view_readonly_properties_cache[source.source]
            elif isinstance(source.source, ContainerId):
                if source.source in constants.READONLY_CONTAINER_PROPERTIES:
                    readonly_properties = constants.READONLY_CONTAINER_PROPERTIES[source.source]

            source.properties = {k: v for k, v in source.properties.items() if k not in readonly_properties}

    def stream_data(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[Page]:
        if isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            chunk = InstanceList([])
            total = 0
            for instance in iterate_instances(client=self.client, **selector.as_filter_args()):
                if limit is not None and total >= limit:
                    break
                total += 1
                chunk.append(instance)
                if len(chunk) >= self.CHUNK_SIZE:
                    yield Page(worker_id="main", items=chunk)
                    chunk = InstanceList([])
            if chunk:
                yield Page(worker_id="main", items=chunk)
        elif isinstance(selector, InstanceFileSelector):
            for node_chunk in chunker_sequence(selector.node_ids, self.CHUNK_SIZE):
                yield Page(
                    worker_id="main",
                    items=InstanceList(self.client.data_modeling.instances.retrieve(nodes=node_chunk).nodes),
                )
            for edge_chunk in chunker_sequence(selector.edge_ids, self.CHUNK_SIZE):
                yield Page(
                    worker_id="main",
                    items=InstanceList(self.client.data_modeling.instances.retrieve(edges=edge_chunk).edges),
                )
        else:
            raise NotImplementedError()

    def download_ids(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[list[InstanceId]]:
        if isinstance(selector, InstanceFileSelector) and selector.validate_instance is False:
            instances_to_yield = selector.instance_ids
            if limit is not None:
                instances_to_yield = instances_to_yield[:limit]
            yield from chunker_sequence(instances_to_yield, self.CHUNK_SIZE)
        else:
            yield from ([instance.as_id() for instance in chunk.items] for chunk in self.stream_data(selector, limit))

    def count(self, selector: InstanceSelector) -> int | None:
        if isinstance(selector, InstanceViewSelector) or (
            isinstance(selector, InstanceSpaceSelector) and selector.view
        ):
            result = self.client.data_modeling.instances.aggregate(
                # MyPy do not understand that selector.view is always defined here.
                view=selector.view.as_id(),  # type: ignore[union-attr]
                aggregates=Count("externalId"),
                instance_type=selector.instance_type,
                space=selector.get_instance_spaces(),
            )
            return int(result.value or 0)
        elif isinstance(selector, InstanceSpaceSelector):
            statistics = self.client.data_modeling.statistics.spaces.retrieve(space=selector.instance_space)
            if statistics is None:
                return None
            if selector.instance_type == "node":
                return statistics.nodes
            elif selector.instance_type == "edge":
                return statistics.edges
            # This should never happen due to validation in the selector.
            raise ValueError(f"Unknown instance type {selector.instance_type!r}")
        elif isinstance(selector, InstanceFileSelector):
            return len(selector.instance_ids)
        raise NotImplementedError()

    def data_to_json_chunk(
        self, data_chunk: Sequence[Instance], selector: InstanceSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        return [instance.as_write().dump(camel_case=True) for instance in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceApply:
        # There is a bug in the SDK where InstanceApply._load turns all keys to snake_case.
        # So we cannot use InstanceApplyList._load here.
        instance_type = item_json.get("instanceType")
        item_to_load = dict(item_json)
        if self._remove_existing_version and "existingVersion" in item_to_load:
            del item_to_load["existingVersion"]
        instance: InstanceApply
        if instance_type == "node":
            instance = NodeApply._load(item_to_load, cognite_client=self.client)
        elif instance_type == "edge":
            instance = EdgeApply._load(item_to_load, cognite_client=self.client)
        else:
            raise ValueError(f"Unknown instance type {instance_type!r}")
        # Filter out read-only properties if applicable
        self._filter_readonly_properties(instance)
        return instance

    def configurations(self, selector: InstanceSelector) -> Iterable[StorageIOConfig]:
        if not isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            return
        spaces = list(set((selector.get_instance_spaces() or []) + (selector.get_schema_spaces() or [])))
        if not spaces:
            return
        space_crud = SpaceCRUD.create_loader(self.client)
        retrieved_spaces = space_crud.retrieve(spaces)
        retrieved_spaces = SpaceList([space for space in retrieved_spaces if not space.is_global])
        if not retrieved_spaces:
            return
        for space in retrieved_spaces:
            yield StorageIOConfig(
                kind=SpaceCRUD.kind,
                folder_name=SpaceCRUD.folder_name,
                value=space_crud.dump_resource(space),
                filename=sanitize_filename(space.space),
            )
        if not selector.view:
            return
        view_id = selector.view.as_id()
        view_crud = ViewCRUD(self.client, None, None, topological_sort_implements=True)
        views = view_crud.retrieve([view_id])
        views = ViewList([view for view in views if not view.is_global])
        if not views:
            return
        for view in views:
            filename = f"{view.space}_{view.external_id}"
            if view.version is not None:
                filename += f"_{view.version}"
            yield StorageIOConfig(
                kind=ViewCRUD.kind,
                folder_name=ViewCRUD.folder_name,
                value=view_crud.dump_resource(view),
                filename=sanitize_filename(filename),
            )
        container_ids = list({container for view in views for container in view.referenced_containers() or []})
        if not container_ids:
            return
        container_crud = ContainerCRUD.create_loader(self.client)
        containers = container_crud.retrieve(container_ids)
        containers = ContainerList([container for container in containers if not container.is_global])
        if not containers:
            return
        for container in containers:
            yield StorageIOConfig(
                kind=ContainerCRUD.kind,
                folder_name=ContainerCRUD.folder_name,
                value=container_crud.dump_resource(container),
                filename=sanitize_filename(f"{container.space}_{container.external_id}"),
            )



from abc import ABC, abstractmethod
from collections.abc import Hashable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar

from rich.markup import escape

from cognite_toolkit._cdf_tk.constants import HINT_LEAD_TEXT

from .base import SeverityLevel, ToolkitWarning


@dataclass(frozen=True)
class FileReadWarning(ToolkitWarning, ABC):
    severity: ClassVar[SeverityLevel]
    filepath: Path

    def group_key(self) -> tuple[Any, ...]:
        return (self.filepath,)

    def group_header(self) -> str:
        return f"    In File {self.filepath.as_posix()!r}"

    def __str__(self) -> str:
        return self.get_message()


@dataclass(frozen=True)
class FileExistsWarning(FileReadWarning):
    severity = SeverityLevel.MEDIUM
    extra_message: str = ""

    def get_message(self) -> str:
        msg = f"{type(self).__name__}: The file {self.filepath} already exists."
        if self.extra_message:
            msg += f" {self.extra_message}"
        return msg


@dataclass(frozen=True)
class IdentifiedResourceFileReadWarning(FileReadWarning, ABC):
    id_value: str
    id_name: str


@dataclass(frozen=True)
class YAMLFileWarning(FileReadWarning, ABC):
    def __post_init__(self) -> None:
        if self.filepath.suffix not in {".yaml", ".yml"}:
            raise ValueError(f"Expected a YAML file, got {self.filepath.suffix}.")


@dataclass(frozen=True)
class YAMLFileWithElementWarning(YAMLFileWarning, ABC):
    # None is a dictionary, number is a list
    element_no: int | None
    path: tuple[str | int, ...]

    @property
    def _location(self) -> str:
        if self.element_no is not None:
            value = f" in entry {self.element_no} "
        else:
            value = ""
        if len(self.path) <= 1:
            return f"{value}"
        else:
            return f"{value} in section {self.path!r}"


@dataclass(frozen=True)
class DuplicatedItemWarning(YAMLFileWarning):
    severity = SeverityLevel.MEDIUM
    identifier: Hashable
    first_location: Path

    def get_message(self) -> str:
        return (
            f"{type(self).__name__}: Duplicated item with identifier "
            f"{self.identifier!r} first seen in {self.first_location.name!r}."
        )


@dataclass(frozen=True)
class UnknownResourceTypeWarning(YAMLFileWarning):
    severity = SeverityLevel.MEDIUM

    suggestion: str | None

    def get_message(self) -> str:
        msg = f"{type(self).__name__}: File {self.filepath.as_posix()!r}."
        if self.suggestion:
            msg += f"\n[blue]HINT: [/blue]{self.suggestion}"
        return msg


@dataclass(frozen=True)
class UnusedParameterWarning(YAMLFileWithElementWarning):
    severity = SeverityLevel.LOW
    actual: str

    def get_message(self) -> str:
        message = f"{type(self).__name__}: Parameter {self.actual!r} is not used{self._location}."
        if self.actual == "dataSetId" or self.actual == "data_set_id":
            message += f" {HINT_LEAD_TEXT}The parameter 'dataSetExternalId' should be used instead."
        return message


@dataclass(frozen=True)
class UnresolvedVariableWarning(FileReadWarning):
    severity = SeverityLevel.HIGH
    variable: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Variable {self.variable!r} is not resolved in {self.filepath}."


@dataclass(frozen=True)
class ResourceMissingIdentifierWarning(YAMLFileWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    message: ClassVar[str] = "The resource is missing an identifier:"
    resource: str
    ext_id_type: str

    def get_message(self) -> str:
        message = f"The {self.resource} {self.filepath} is missing the {self.ext_id_type} field(s)."
        return message


@dataclass(frozen=True)
class NamingConventionWarning(YAMLFileWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: ClassVar[str] = "The naming convention is not followed:"
    resource: str
    ext_id_type: str
    external_id: str

    @property
    @abstractmethod
    def recommendation(self) -> str:
        raise NotImplementedError()

    def get_message(self) -> str:
        message = (
            f"The {self.ext_id_type} identifier [bold]{self.external_id!r}[/bold] of the resource {self.resource} "
            f"does not follow the recommended naming convention {self.recommendation}"
        )
        return message


@dataclass(frozen=True)
class PrefixConventionWarning(NamingConventionWarning):
    prefix: str

    @property
    def recommendation(self) -> str:
        return f"of prefixing with {self.prefix!r}."


@dataclass(frozen=True)
class NamespacingConventionWarning(NamingConventionWarning):
    namespace: str

    @property
    def recommendation(self) -> str:
        return f"of using {self.namespace!r} as separator."


@dataclass(frozen=True)
class CaseTypoWarning(UnusedParameterWarning):
    expected: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Got {self.actual!r}. Did you mean {self.expected!r}?{self._location}."


@dataclass(frozen=True)
class MissingRequiredParameterWarning(YAMLFileWithElementWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    expected: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Missing required parameter {self.expected!r}{self._location}."


@dataclass(frozen=True)
class MissingRequiredIdentifierWarning(YAMLFileWithElementWarning):
    expected: tuple[str, ...]

    def get_message(self) -> str:
        return f"{type(self).__name__}: Missing required identifier {self.expected!r}{self._location}."


@dataclass(frozen=True)
class MissingReferencedWarning(YAMLFileWithElementWarning):
    message: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: {self.message}."


@dataclass(frozen=True)
class TemplateVariableWarning(IdentifiedResourceFileReadWarning):
    path: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Variable {self.id_name!r} has value {self.id_value!r} in file: {self.filepath.name}. Did you forget to change it?"


@dataclass(frozen=True)
class DataSetMissingWarning(IdentifiedResourceFileReadWarning):
    severity = SeverityLevel.MEDIUM
    resource_name: str

    def get_message(self) -> str:
        # Avoid circular import
        from cognite_toolkit._cdf_tk.cruds import TransformationCRUD

        if self.filepath.parent.name == TransformationCRUD.folder_name:
            return f"{type(self).__name__}: It is recommended to use a data set if source or destination can be scoped with a data set. If not, ignore this warning."
        else:
            return f"{type(self).__name__}: It is recommended that you set dataSetExternalId for {self.resource_name}. This is missing in {self.filepath.name}. Did you forget to add it?"


@dataclass(frozen=True)
class SourceFileModifiedWarning(FileReadWarning):
    severity = SeverityLevel.ERROR

    def get_message(self) -> str:
        message = (
            f"{type(self).__name__}: The source file {self.filepath} has been modified since the last build. "
            "Please rebuild the project."
        )
        return message


@dataclass(frozen=True)
class MissingFileWarning(FileReadWarning):
    severity = SeverityLevel.MEDIUM
    attempted_check: str

    def get_message(self) -> str:
        message = f"{type(self).__name__}: The file {self.filepath} is missing. Cannot verify {self.attempted_check}."
        return message


@dataclass(frozen=True)
class EnvironmentVariableMissingWarning(FileReadWarning):
    severity = SeverityLevel.HIGH
    variables: frozenset[str]
    identifiers: frozenset[Hashable] | None = None

    def get_message(self) -> str:
        from cognite_toolkit._cdf_tk.utils import humanize_collection

        suffix = "s are" if len(self.variables) > 1 else " is"
        variables = humanize_collection(self.variables, sort=True, bind_word="and")
        return f"The environment variable{suffix} missing: {variables}"


@dataclass(frozen=True)
class StreamlitRequirementsWarning(FileReadWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.MEDIUM
    dependencies: str | list[str]

    def get_message(self) -> str:
        if isinstance(self.dependencies, str):
            return f"Missing recommended dependencies in requirements.txt: {self.dependencies}"
        return f"Missing dependencies in requirements.txt: {', '.join(self.dependencies)}"


@dataclass(frozen=True)
class ResourceFormatWarning(FileReadWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    errors: tuple[str, ...]
    text: str | None = None

    def get_message(self) -> str:
        sep = "\n     * "
        errors = sep.join(map(escape, self.errors))
        s = "s" if len(self.errors) > 1 else ""
        text = self.text or ""
        return f"{type(self).__name__} {text}{len(self.errors)} error{s}:{sep}{errors}"



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.workflow_version import WorkflowVersionYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_workflow_version_test_cases() -> Iterable:
    # Missing required top-level fields
    yield pytest.param(
        {},
        {
            "Missing required field: 'workflowExternalId'",
            "Missing required field: 'version'",
            "Missing required field: 'workflowDefinition'",
        },
        id="Missing all required top-level fields",
    )
    # Extra/unknown field at top level
    yield pytest.param(
        {
            "workflowExternalId": "wf1",
            "version": "v1",
            "workflowDefinition": {"description": "desc", "tasks": []},
            "foo": 123,
        },
        {"Unused field: 'foo'"},
        id="Extra field at top level",
    )
    # Missing required fields in task definition
    yield pytest.param(
        {"workflowExternalId": "wf1", "version": "v1", "workflowDefinition": {"description": "desc", "tasks": [{}]}},
        {
            "In workflowDefinition.tasks[1] missing required field: 'externalId'",
            "In workflowDefinition.tasks[1] missing required field: 'parameters'",
            "In workflowDefinition.tasks[1] missing required field: 'type'",
        },
        id="Missing required fields in task definition",
    )
    # Wrong type for retries in task definition
    yield pytest.param(
        {
            "workflowExternalId": "wf1",
            "version": "v1",
            "workflowDefinition": {
                "description": "desc",
                "tasks": [{"externalId": "t1", "type": "function", "parameters": {}, "retries": "notAnInt"}],
            },
        },
        {"In workflowDefinition.tasks[1].retries input should be a valid integer. Got 'notAnInt' of type str."},
        id="Wrong type for retries",
    )
    # Invalid enum value for on_failure
    yield pytest.param(
        {
            "workflowExternalId": "wf1",
            "version": "v1",
            "workflowDefinition": {
                "description": "desc",
                "tasks": [{"externalId": "t1", "type": "function", "parameters": {}, "onFailure": "notAValidValue"}],
            },
        },
        {
            "In workflowDefinition.tasks[1].onFailure input should be 'abortWorkflow' or "
            "'skipTask'. Got 'notAValidValue'."
        },
        id="Invalid enum value for onFailure",
    )


class TestWorkflowVersionYAML:
    @pytest.mark.parametrize("data", list(find_resources("WorkflowVersion")))
    def test_load_valid_workflow_file(self, data: dict[str, object]) -> None:
        loaded = WorkflowVersionYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_workflow_version_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, WorkflowVersionYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from cognite.client.data_classes import WorkflowUpsert
from pydantic import Field

from .base import ToolkitResource


class WorkflowYAML(ToolkitResource):
    _cdf_resource = WorkflowUpsert
    external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project."
        " No trailing or leading whitespace and no null characters allowed.",
    )
    description: str | None = Field(None, max_length=500)
    data_set_external_id: str | None = None



from pydantic import Field

from .authentication import AuthenticationClientIdSecret
from .base import ToolkitResource


class HostedExtractorDestinationYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    credentials: AuthenticationClientIdSecret | None = Field(
        None, description="Credentials for authenticating towards CDF using a CDF session."
    )
    target_data_set_external_id: str | None = Field(
        None,
        description="The external ID of the target data set where the extractor will write data.",
        max_length=255,
    )



import contextlib
import io
from pathlib import Path
from unittest.mock import MagicMock, patch

from cognite.client.data_classes.capabilities import FilesAcl, FunctionsAcl
from cognite.client.data_classes.workflows import WorkflowScheduledTriggerRule, WorkflowTrigger

from cognite_toolkit._cdf_tk.cruds import FunctionCRUD, ResourceWorker, WorkflowTriggerCRUD
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestResourceWorker:
    def test_mask_sensitive_data(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        toolkit_client_approval.append(
            WorkflowTrigger,
            WorkflowTrigger(
                "my_trigger",
                WorkflowScheduledTriggerRule(cron_expression="* * * * *"),
                "my_workflow",
                "v1",
                metadata={
                    WorkflowTriggerCRUD._MetadataKey.secret_hash: "outdated-hash",
                },
            ),
        )
        loader = WorkflowTriggerCRUD.create_loader(toolkit_client_approval.mock_client)

        worker = ResourceWorker(loader, "deploy")
        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = """externalId: my_trigger
triggerRule:
  triggerType: schedule
  cronExpression: '* * * * *'
workflowExternalId: my_workflow
workflowVersion: v1
authentication:
  clientId: my_client_id
  clientSecret: my_super_secret_42
"""
        output_capture = io.StringIO()
        with contextlib.redirect_stdout(output_capture):
            _ = worker.prepare_resources([local_file], environment_variables={}, is_dry_run=False, verbose=True)

        terminal_output = output_capture.getvalue()
        assert "my_super_secret_42" not in terminal_output

    def test_worker_uses_function_capabilities(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        # This test verifies that the ResourceWorker uses function-specific capabilities
        # for FunctionLoader rather than generic capabilities
        with patch(
            "cognite_toolkit._cdf_tk.cruds._resource_cruds.function.FunctionCRUD.load_resource_file"
        ) as mock_load_resource_file:
            mock_authorization = toolkit_client_approval.mock_client.verify.authorization
            mock_authorization.return_value = []
            mock_load_resource_file.return_value = [
                {
                    "externalId": "my_function",
                    "name": "My Function",
                    "fileId": 123,
                }
            ]

            loader = FunctionCRUD.create_loader(toolkit_client_approval.mock_client, None)
            loader.data_set_id_by_external_id = {"my_function": 789}

            local_file = MagicMock(spec=Path)
            local_file.parent.name = FunctionCRUD.folder_name

            worker = ResourceWorker(loader, "deploy")
            local_by_id = worker.load_resources([local_file], None, False)
            worker.validate_access(local_by_id, is_dry_run=False)
            mock_authorization.assert_called_once()

            capabilities_arg = mock_authorization.call_args[0][0]
            assert len(capabilities_arg) == 2
            assert isinstance(capabilities_arg[0], FunctionsAcl)
            assert isinstance(capabilities_arg[1], FilesAcl)
            assert isinstance(capabilities_arg[1].scope, FilesAcl.Scope.DataSet)
            assert capabilities_arg[1].scope.ids == [789]



import pytest

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.streams import StreamRequest, StreamResponse
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.mark.skip("Soft deletes limits stop this from running frequently")
class TestStreamsAPI:
    def test_create_list_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        stream = StreamRequest.model_validate(
            {
                "externalId": f"test_stream_{RUN_UNIQUE_ID}".lower(),
                "settings": {"template": {"name": "ImmutableTestStream"}},
            }
        )

        # create stream
        created_list = toolkit_client.streams.create([stream])
        assert len(created_list) >= 1
        assert all(isinstance(item, StreamResponse) for item in created_list)
        assert any(item.external_id == stream.external_id for item in created_list)

        # list streams
        all_streams = toolkit_client.streams.list()
        assert len(all_streams) >= 1
        assert any(item.external_id == stream.external_id for item in all_streams)

        # retrieve stream
        retrieved = toolkit_client.streams.retrieve(stream.external_id)
        assert isinstance(retrieved, StreamResponse)
        assert retrieved.external_id == stream.external_id
        assert retrieved.created_time > 0
        assert retrieved.created_from_template == "ImmutableTestStream"

        # delete stream
        toolkit_client.streams.delete(stream.external_id)

        # list streams after delete
        all_streams_after_delete = toolkit_client.streams.list()
        assert len(all_streams_after_delete) >= 0
        assert not any(item.external_id == stream.external_id for item in all_streams_after_delete or [])



from pathlib import Path

import pytest
from cognite.client.data_classes import Asset, Transformation
from openpyxl import load_workbook

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import (
    ProfileAssetCentricCommand,
    ProfileAssetCommand,
    ProfileRawCommand,
    ProfileTransformationCommand,
)
from tests.test_integration.constants import (
    ASSET_COUNT,
    ASSET_DATASET,
    ASSET_TABLE,
    EVENT_COUNT,
    EVENT_DATASET,
    EVENT_TABLE,
    FILE_COUNT,
    FILE_DATASET,
    FILE_TABLE,
    SEQUENCE_COUNT,
    SEQUENCE_DATASET,
    SEQUENCE_TABLE,
    TIMESERIES_COUNT,
    TIMESERIES_DATASET,
    TIMESERIES_TABLE,
)


class TestProfileAssetCommand:
    @pytest.mark.usefixtures("disable_throttler")
    def test_profile_asset_hierarchy(
        self,
        toolkit_client: ToolkitClient,
        aggregator_root_asset: Asset,
        aggregator_raw_db: str,
        aggregator_assets: Transformation,
        aggregator_events: Transformation,
        aggregator_files: Transformation,
        aggregator_time_series: Transformation,
        aggregator_sequences: Transformation,
    ) -> None:
        profile_row_limit = TIMESERIES_COUNT - 1  # Force fallaback to using the preview endpoint for row count
        results = ProfileAssetCommand().assets(
            toolkit_client, aggregator_root_asset.external_id, profile_row_limit=profile_row_limit
        )
        for row in results:
            for cell in row.values():
                if "CogniteAPIError" in str(cell):
                    pytest.skip("Skipping test due to intermediate CogniteAPI error.")

        columns = ProfileAssetCommand.Columns
        assert (
            results
            == [
                {
                    columns.Resource: resource,
                    columns.Count: count,
                    columns.DataSets: dataset,
                    columns.DataSetCount: count,
                    columns.Transformations: f"{transformation.name} ({transformation.external_id})",
                    columns.RawTable: f"{aggregator_raw_db}.{raw_table}",
                    columns.RowCount: row_count,
                    columns.ColumnCount: column_count,
                }
                for resource, count, dataset, transformation, raw_table, row_count, column_count in [
                    (
                        "Assets",
                        ASSET_COUNT,
                        ASSET_DATASET,
                        aggregator_assets,
                        ASSET_TABLE,
                        ASSET_COUNT - 1,
                        4,
                    ),  # -1 root asset is not in the table
                    (
                        "Events",
                        EVENT_COUNT,
                        EVENT_DATASET,
                        aggregator_events,
                        EVENT_TABLE,
                        EVENT_COUNT,
                        5,
                    ),
                    (
                        "Files",
                        FILE_COUNT,
                        FILE_DATASET,
                        aggregator_files,
                        FILE_TABLE,
                        FILE_COUNT,
                        4,
                    ),
                    (
                        "TimeSeries",
                        TIMESERIES_COUNT,
                        TIMESERIES_DATASET,
                        aggregator_time_series,
                        TIMESERIES_TABLE,
                        TIMESERIES_COUNT,
                        "≥5",  # Force fallback to using the preview endpoint for row count, then we get "≥5" as the column count
                    ),
                    (
                        "Sequences",
                        SEQUENCE_COUNT,
                        SEQUENCE_DATASET,
                        aggregator_sequences,
                        SEQUENCE_TABLE,
                        SEQUENCE_COUNT,
                        4,
                    ),
                ]
            ]
        )


class TestProfileAssetCentric:
    def test_profile_assent_centric(self, toolkit_client: ToolkitClient, monkeypatch, tmp_path: Path) -> None:
        output_spreadsheet = tmp_path / "asset_centric_profile.xlsx"
        cmd = ProfileAssetCentricCommand(output_spreadsheet)
        results = cmd.asset_centric(toolkit_client, select_all=True, verbose=False)

        assert len(results) == 7
        assert {item["Resource"] for item in results} == {
            "Assets",
            "Events",
            "Files",
            "TimeSeries",
            "Sequences",
            "Relationships",
            "Labels",
        }
        total_count = sum(item["Count"] for item in results)
        assert total_count > 0
        total_metadata_count = 0
        for item in results:
            metadata_count = item.get(ProfileAssetCentricCommand.Columns.MetadataKeyCount)
            if not metadata_count:
                continue
            total_metadata_count += metadata_count
        assert total_metadata_count > 0

        workbook = load_workbook(output_spreadsheet)
        assert cmd.table_title in workbook.sheetnames
        sheet = workbook[cmd.table_title]
        assert sheet.max_row == len(results) + 1  # +1 for header row

    def test_profile_asset_centric_hierarchy(
        self, toolkit_client: ToolkitClient, aggregator_root_asset: Asset, tmp_path: Path
    ) -> None:
        output_spreadsheet = tmp_path / "asset_centric_hierarchy_profile.xlsx"
        cmd = ProfileAssetCentricCommand(silent=True, output_spreadsheet=output_spreadsheet)
        results = cmd.asset_centric(toolkit_client, hierarchy=aggregator_root_asset.external_id, verbose=False)
        actual = [{"Resource": row["Resource"], "Count": row["Count"]} for row in results]
        expected = [
            {"Resource": "Assets", "Count": ASSET_COUNT},
            {"Resource": "Events", "Count": EVENT_COUNT},
            {"Resource": "Files", "Count": FILE_COUNT},
            {"Resource": "TimeSeries", "Count": TIMESERIES_COUNT},
            {"Resource": "Sequences", "Count": SEQUENCE_COUNT},
        ]
        assert actual == expected, f"Expected {expected}, but got {actual}"

        workbook = load_workbook(output_spreadsheet)
        assert "Assets" in workbook.sheetnames
        sheet = workbook["Assets"]
        assert [col.value for (col, *_) in sheet.columns] == ["Metadata Key", "Count"]


class TestProfileTransformationCommand:
    def test_profile_transformation(
        self, toolkit_client: ToolkitClient, aggregator_assets: Transformation, aggregator_raw_db: str
    ) -> None:
        results = ProfileTransformationCommand().transformation(toolkit_client, "assets")
        columns = ProfileTransformationCommand.Columns
        search_rows = [row for row in results if row[columns.Transformation] == aggregator_assets.name]
        assert len(search_rows) == 1, "Expected exactly one row for the transformation"
        actual_row = search_rows[0]
        assert actual_row == {
            columns.Transformation: aggregator_assets.name,
            columns.Source: f"{aggregator_raw_db}.{ASSET_TABLE}",
            columns.DestinationColumns: "name, externalId, dataSetId, parentExternalId",
            columns.Destination: "assets",
            columns.ConflictMode: aggregator_assets.conflict_mode,
            columns.IsPaused: "No schedule",
        }


class TestProfileRawCommand:
    def test_profile_raw(
        self, toolkit_client: ToolkitClient, aggregator_raw_db: str, aggregator_assets: Transformation
    ) -> None:
        transformation = aggregator_assets
        results = ProfileRawCommand().raw(toolkit_client, destination_type="assets", verbose=False)
        columns = ProfileRawCommand.Columns
        search_rows = [row for row in results if row[columns.RAW] == f"{aggregator_raw_db}.{ASSET_TABLE}"]
        assert len(search_rows) == 1, f"Expected exactly one row for the raw table {ASSET_TABLE}."
        actual_row = search_rows[0]
        assert actual_row == {
            columns.RAW: f"{aggregator_raw_db}.{ASSET_TABLE}",
            columns.Rows: ASSET_COUNT - 1,  # -1 root asset is not in the RAW table.
            columns.Columns: 4,
            columns.Transformation: f"{transformation.name} ({transformation.external_id})",
            columns.Destination: "assets",
            columns.ConflictMode: transformation.conflict_mode,
        }



import sys
from abc import ABC
from typing import Any, Literal, TypeAlias

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

from cognite_toolkit._cdf_tk.client.data_classes.charts_data import ChartData

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

Visibility: TypeAlias = Literal["PUBLIC", "PRIVATE"]


class ChartCore(WriteableCogniteResource["ChartWrite"], ABC):
    """Base class for the Chart data model.

    Args:
        external_id (str): Unique identifier for the chart.
        visibility (Visibility): Visibility of the chart, either 'PUBLIC' or 'PRIVATE'.
        data (ChartData): The data associated with the chart.
    """

    def __init__(self, external_id: str, visibility: Visibility, data: ChartData) -> None:
        self.external_id = external_id
        self.visibility = visibility
        self.data = data

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Convert the chart to a dictionary representation."""
        output = super().dump(camel_case=camel_case)
        output["data"] = self.data.dump(camel_case=camel_case)
        return output


class ChartWrite(ChartCore):
    """A chart that can be written to the CDF.

    Args:
        external_id (str): Unique identifier for the chart.
        visibility (Visibility): Visibility of the chart, either 'PUBLIC' or 'PRIVATE'.
        data (ChartData): The data associated with the chart.

    """

    def as_write(self) -> "ChartWrite":
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            visibility=resource["visibility"],
            data=ChartData._load(resource["data"], cognite_client=cognite_client),
        )


class Chart(ChartCore):
    """A chart that can be read from the CDF.

    Args:
        external_id (str): Unique identifier for the chart.
        created_time (int): Timestamp when the chart was created.
        last_updated_time (int): Timestamp when the chart was last updated.
        visibility (Visibility): Visibility of the chart, either 'PUBLIC' or 'PRIVATE'.
        data (ChartData): The data associated with the chart.
        owner_id (str): The ID of the user who owns the chart.
    """

    def __init__(
        self,
        external_id: str,
        created_time: int,
        last_updated_time: int,
        visibility: Visibility,
        data: ChartData,
        owner_id: str,
    ) -> None:
        super().__init__(external_id, visibility, data)
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.owner_id = owner_id

    def as_write(self) -> ChartWrite:
        return ChartWrite(external_id=self.external_id, visibility=self.visibility, data=self.data)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            visibility=resource["visibility"],
            data=ChartData._load(resource["data"], cognite_client=cognite_client),
            owner_id=resource["ownerId"],
        )


class ChartWriteList(CogniteResourceList):
    _RESOURCE = ChartWrite


class ChartList(WriteableCogniteResourceList[ChartWrite, Chart]):
    _RESOURCE = Chart

    def as_write(self) -> ChartWriteList:
        return ChartWriteList([item.as_write() for item in self])



import json
from collections.abc import Iterator
from pathlib import Path

import pytest
import respx
from cognite.client.data_classes.raw import RowWrite, Table, TableList

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable, RawTableList
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.constants import DATA_RESOURCE_DIR
from cognite_toolkit._cdf_tk.cruds import RawTableCRUD
from cognite_toolkit._cdf_tk.storageio import RawIO
from cognite_toolkit._cdf_tk.storageio.selectors import RawTableSelector, SelectedTable
from cognite_toolkit._cdf_tk.utils.fileio import NDJsonWriter, Uncompressed


@pytest.fixture
def raw_json_directory(tmp_path: Path) -> Path:
    """Fixture to create a temporary folder with a sample NDJSON file."""
    configfile = tmp_path / DATA_RESOURCE_DIR / RawTableCRUD.folder_name / f"test_table.{RawTableCRUD.kind}.yaml"
    configfile.parent.mkdir(parents=True, exist_ok=True)
    table = RawTable(db_name="test_db", table_name="test_table")
    configfile.write_text(table.dump_yaml())
    with NDJsonWriter(tmp_path, RawIO.KIND, Uncompressed) as writer:
        writer.write_chunks(
            [
                RowWrite(
                    key=f"key{i}",
                    columns={
                        "column1": f"value{i}",
                        "column2": i,
                        "column3": i % 2 == 0,
                    },
                ).dump()
                for i in range(1_000)
            ],
            filestem="test_table",
        )

    selector = RawTableSelector(
        table=SelectedTable(db_name=table.db_name, table_name=table.table_name), type="rawTable"
    )
    selector.dump_to_file(tmp_path)
    return tmp_path


@pytest.fixture
def raw_csv_directory(tmp_path: Path) -> Path:
    """Fixture to create a temporary folder with a sample CSV file."""
    configfile = tmp_path / DATA_RESOURCE_DIR / RawTableCRUD.folder_name / f"test_table.{RawTableCRUD.kind}.yaml"
    configfile.parent.mkdir(parents=True, exist_ok=True)
    table = RawTable(db_name="test_db", table_name="test_table")
    configfile.write_text(table.dump_yaml())
    csv_file = tmp_path / f"test_table.{RawIO.KIND}.csv"
    with csv_file.open("w") as f:
        f.write("index,column1,column2,column3\n")
        for i in range(1, 1001):
            f.write(f"key{i},value{i},{i},{i % 2 == 0}\n")

    selector = RawTableSelector(
        table=SelectedTable(db_name=table.db_name, table_name=table.table_name), type="rawTable", key="index"
    )
    selector.dump_to_file(tmp_path)
    return tmp_path


@pytest.fixture
def raw_mock_client(
    toolkit_config: ToolkitClientConfig,
    respx_mock: respx.MockRouter,
    monkeypatch: pytest.MonkeyPatch,
) -> Iterator[tuple[ToolkitClient, respx.MockRouter]]:
    config = toolkit_config
    insert_url = config.create_api_url("/raw/dbs/test_db/tables/test_table/rows")
    respx_mock.post(url=insert_url).respond(status_code=200)
    monkeypatch.setenv("CDF_CLUSTER", config.cdf_cluster)
    monkeypatch.setenv("CDF_PROJECT", config.project)
    with monkeypatch_toolkit_client() as client:
        client.verify.authorization.return_value = []
        client.raw.tables.list.return_value = RawTableList([])
        client.raw.tables.create.return_value = TableList([Table(name="test_table")])
        client.config = config
        yield client, respx_mock


class TestUploadCommand:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_upload_raw_rows_from_ndjson(
        self,
        raw_mock_client: tuple[ToolkitClient, respx.MockRouter],
        raw_json_directory: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        client, respx_mock = raw_mock_client

        cmd = UploadCommand(silent=True, skip_tracking=True)
        cmd.upload(raw_json_directory, client, deploy_resources=True, dry_run=False, verbose=False)

        self.assert_raw_rows_uploaded(client, respx_mock)

    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_upload_raw_rows_from_csv(
        self,
        raw_mock_client: tuple[ToolkitClient, respx.MockRouter],
        raw_csv_directory: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        client, respx_mock = raw_mock_client

        cmd = UploadCommand(silent=True, skip_tracking=True)
        cmd.upload(raw_csv_directory, client, deploy_resources=True, dry_run=False, verbose=False)

        self.assert_raw_rows_uploaded(client, respx_mock)

    def assert_raw_rows_uploaded(self, client: ToolkitClient, respx_mock: respx.MockRouter) -> None:
        assert len(respx_mock.calls) == 1
        call = respx_mock.calls[0]
        assert str(call.request.url).endswith("/raw/dbs/test_db/tables/test_table/rows")
        body = json.loads(call.request.content)["items"]
        assert isinstance(body, list)
        assert len(body) == 1_000

        client.raw.tables.create.assert_called_once()
        _, table_kwargs = client.raw.tables.create.call_args
        assert table_kwargs["db_name"] == "test_db"
        assert table_kwargs["name"] == ["test_table"]

    def test_upload_raw_rows_dry_run(
        self, toolkit_config: ToolkitClientConfig, raw_json_directory: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        config = toolkit_config
        cmd = UploadCommand(silent=True, skip_tracking=True)
        monkeypatch.setenv("CDF_CLUSTER", config.cdf_cluster)
        monkeypatch.setenv("CDF_PROJECT", config.project)
        with monkeypatch_toolkit_client() as client:
            client.verify.authorization.return_value = []
            cmd.upload(raw_json_directory, client, deploy_resources=True, dry_run=True, verbose=False)

            client.raw.rows.insert.assert_not_called()
            client.raw.databases.create.assert_not_called()
            client.raw.tables.create.assert_not_called()



import json
from unittest.mock import MagicMock, patch

import pytest
import respx
from cognite.client import global_config
from cognite.client.data_classes import FunctionWrite
from httpx import Response
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ToolkitAPIError


class TestExtendedFunctionsAPI:
    def test_create_function_200(self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions")
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")

        respx_mock.post(url).mock(
            return_value=Response(
                status_code=200,
                json={"items": [{"externalId": "test_function", "name": "test_function", "createdTime": 42}]},
            )
        )
        result = client.functions.create_with_429_retry(fun)

        assert result.external_id == "test_function"
        assert result.name == "test_function"
        assert result.created_time == 42

    def test_create_function_429_succeed(
        self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig
    ) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions")
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)

        with patch(f"{HTTPClient.__module__}.time"):
            # Add multiple 429 responses followed by a success response
            responses = [
                Response(status_code=429, json={"error": "Too many requests"}, headers={"Retry-After": "42"})
                for _ in range(global_config.max_retries - 1)
            ]
            responses.append(
                Response(
                    status_code=200,
                    json={"items": [{"externalId": "test_function", "name": "test_function", "createdTime": 42}]},
                )
            )
            respx_mock.post(url).mock(side_effect=responses)

            result = client.functions.create_with_429_retry(fun, console=console)
        assert result.external_id == "test_function"
        assert console.print.call_count == global_config.max_retries - 1
        assert console.print.call_args.args[1] == (
            "Rate limit exceeded for the '/functions' endpoint. Retrying after 42.0 seconds."
        )

    def test_create_function_429_exceed_max_retries(
        self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig
    ) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions")
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)

        with patch(f"{HTTPClient.__module__}.time"):
            # Mock to always return 429 responses
            respx_mock.post(url).mock(
                return_value=Response(
                    status_code=429, json={"error": "Too many requests"}, headers={"Retry-After": "42"}
                )
            )

            with pytest.raises(ToolkitAPIError) as exc_info:
                client.functions.create_with_429_retry(fun, console=console)
        assert console.print.call_count == global_config.max_retries
        assert "Too many requests" in str(exc_info.value)
        assert (
            "Rate limit exceeded for the '/functions' endpoint. Retrying after 42.0 seconds."
        ) in console.print.call_args.args[1]

    def test_create_function_429_invalid_retry_after(
        self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig
    ) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions")
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)

        with patch(f"{HTTPClient.__module__}.time"):
            respx_mock.post(url).mock(
                return_value=Response(
                    status_code=429, json={"error": "Too many requests"}, headers={"Retry-After": "invalid"}
                )
            )

            with pytest.raises(ToolkitAPIError) as exc_info:
                client.functions.create_with_429_retry(fun, console=console)
        assert console.print.call_count == 0
        assert "Too many requests" in str(exc_info.value)

    def test_create_function_401(self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions")
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")

        from cognite_toolkit._cdf_tk.utils.http_client import ToolkitAPIError

        respx_mock.post(url).mock(return_value=Response(status_code=401, json={"error": "Unauthorized"}))

        with pytest.raises(ToolkitAPIError) as exc_info:
            client.functions.create_with_429_retry(fun)

        assert "401" in str(exc_info.value)
        assert "Unauthorized" in str(exc_info.value)
        # Verify that only one request was made
        assert len(respx_mock.calls) == 1
        assert respx_mock.calls[0].request.url == url
        assert respx_mock.calls[0].request.method == "POST"

    def test_create_function_invalid_json_float(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function", cpu=float("inf"))

        with pytest.raises(ToolkitAPIError) as exc_info:
            client.functions.create_with_429_retry(fun)

        assert "Out of range float values are not JSON compliant" in str(exc_info.value)

    def test_create_function_invalid_json(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        fun = FunctionWrite(
            name="test_function",
            file_id=123,
            external_id="test_function",
            cpu=43j,  # Complex number, which is not JSON serializable
        )
        with pytest.raises(ToolkitAPIError) as exc_info:
            client.functions.create_with_429_retry(fun)

        assert "Object 43j of type <class 'complex'> can't be serialized by the JSON encoder" in str(exc_info.value)

    @pytest.mark.usefixtures("disable_gzip")
    def test_delete_function_200(self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions/delete")

        respx_mock.post(url).mock(return_value=Response(status_code=200, json={}))

        # Should not raise any exception
        client.functions.delete_with_429_retry(["test_function"], ignore_unknown_ids=True)

        assert len(respx_mock.calls) == 1
        call = respx_mock.calls[0]
        assert call.request.url == url
        assert json.loads(call.request.content) == {
            "items": [{"externalId": "test_function"}],
            "ignoreUnknownIds": True,
        }

    def test_delete_function_429_succeed(
        self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig
    ) -> None:
        config = toolkit_config
        client = ToolkitClient(config=config, enable_set_pending_ids=True)
        url = config.create_api_url("/functions/delete")
        console = MagicMock(spec=Console)

        with patch(f"{HTTPClient.__module__}.time"):
            # Add multiple 429 responses followed by a success response
            responses = [
                Response(status_code=429, json={"error": "Too many requests"}, headers={"Retry-After": "42"})
                for _ in range(global_config.max_retries - 1)
            ]
            responses.append(Response(status_code=200, json={}))
            respx_mock.post(url).mock(side_effect=responses)

            # Should not raise any exception
            client.functions.delete_with_429_retry(["test_function"], ignore_unknown_ids=True, console=console)
        assert console.print.call_count == global_config.max_retries - 1
        assert console.print.call_args.args[1] == (
            "Rate limit exceeded for the '/functions/delete' endpoint. Retrying after 42.0 seconds."
        )



import json

import pytest
import respx
from cognite.client.data_classes.raw import Row, RowList

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.storageio import RawIO
from cognite_toolkit._cdf_tk.storageio.selectors import RawTableSelector, SelectedTable
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


@pytest.fixture()
def some_raw_tables() -> RowList:
    """Fixture to provide a sample RowList for testing."""
    return RowList(
        [
            Row(
                key=f"row{i}",
                columns={
                    "column1": f"value1_{i}",
                    "column2": f"value2_{i}",
                    "column3": f"value3_{i}",
                    "column4": {"nested_key": f"nested_value_{i}"},
                },
            )
            for i in range(100)
        ]
    )


class TestRawStorageIO:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_download_upload(
        self, toolkit_config: ToolkitClientConfig, some_raw_tables: RowList, respx_mock: respx.MockRouter
    ) -> None:
        config = toolkit_config
        respx_mock.post(
            config.create_api_url("/raw/dbs/test_db/tables/test_table/rows"),
        ).respond(status_code=200)
        selector = RawTableSelector(table=SelectedTable(db_name="test_db", table_name="test_table"))
        with monkeypatch_toolkit_client() as client:
            client.raw.rows.return_value = chunker(some_raw_tables, 10)
            io = RawIO(client)

            assert io.count(selector) is None

            source = io.stream_data(selector, limit=100)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for chunk in source:
                json_chunk = io.data_to_json_chunk(chunk.items)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                json_chunks.append(json_chunk)

            with HTTPClient(config) as upload_client:
                data_chunks = (io.json_chunk_to_data([("", item) for item in chunk]) for chunk in json_chunks)
                for data_chunk in data_chunks:
                    io.upload_items(data_chunk, upload_client, selector)

            assert respx_mock.calls.call_count == 10  # 100 rows in chunks of 10
            uploaded_rows = []
            for call in respx_mock.calls:
                uploaded_rows.extend(json.loads(call.request.content)["items"])

            assert uploaded_rows == some_raw_tables.as_write().dump()



from pathlib import Path

from cognite.client.data_classes import TimeSeries

from cognite_toolkit._cdf_tk.tk_warnings import (
    DataSetMissingWarning,
)
from cognite_toolkit._cdf_tk.validation import validate_data_set_is_set


def test_validate_data_set_is_set():
    warnings = validate_data_set_is_set(
        {"externalId": "myTimeSeries", "name": "My Time Series"}, TimeSeries, Path("timeseries.yaml")
    )

    assert sorted(warnings) == sorted(
        [DataSetMissingWarning(Path("timeseries.yaml"), "myTimeSeries", "externalId", "TimeSeries")]
    )



from collections.abc import Iterable

import pytest
from cognite.client.data_classes.capabilities import (
    AllScope,
    AssetRootIDScope,
    AssetsAcl,
    Capability,
    DataModelInstancesAcl,
    DataSetScope,
    IDScopeLowerCase,
    ProjectCapability,
    ProjectCapabilityList,
    ProjectsScope,
    SpaceIDScope,
    TimeSeriesAcl,
)
from cognite.client.data_classes.iam import ProjectSpec, TokenInspection

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client


def get_scope_test_cases() -> Iterable:
    yield pytest.param(
        [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([123]))],
        [TimeSeriesAcl.Action.Read],
        [DataSetScope(ids=[123])],
        id="TimeSeriesAcl with DataSet scope",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All()),
            TimeSeriesAcl([TimeSeriesAcl.Action.Write], TimeSeriesAcl.Scope.DataSet([123])),
        ],
        [TimeSeriesAcl.Action.Read],
        [AllScope()],
        id="Prefer all scope over DataSet scope",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All()),
            TimeSeriesAcl([TimeSeriesAcl.Action.Write], TimeSeriesAcl.Scope.DataSet([123])),
        ],
        [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
        [TimeSeriesAcl.Scope.DataSet([123])],
        id="Minimum scope for multiple actions",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1, 2])),
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([2, 3])),
        ],
        [TimeSeriesAcl.Action.Read],
        [DataSetScope(ids=[1, 2, 3])],
        id="Take union of scopes for same action",
    )
    yield pytest.param(
        [DataModelInstancesAcl([DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.All())],
        [DataModelInstancesAcl.Action.Write],
        None,
        id="DataModelInstancesAcl with All scope and no matching action",
    )
    yield pytest.param(
        [DataModelInstancesAcl([DataModelInstancesAcl.Action.Write], DataModelInstancesAcl.Scope.SpaceID(["space1"]))],
        [DataModelInstancesAcl.Action.Write],
        [SpaceIDScope(space_ids=["space1"])],
        id="DataModelInstancesAcl with SpaceID scope",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([123])),
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.ID([789])),
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.AssetRootID([456])),
        ],
        [TimeSeriesAcl.Action.Read],
        [
            AssetRootIDScope(root_ids=[456]),
            DataSetScope(ids=[123]),
            IDScopeLowerCase(ids=[789]),
        ],
        id="Multiple different scopes with AssetsAcl action",
    )
    yield pytest.param(
        [],
        [TimeSeriesAcl.Action.Read],
        None,
        id="No capabilities present",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1])),
            TimeSeriesAcl([TimeSeriesAcl.Action.Write], TimeSeriesAcl.Scope.DataSet([2])),
        ],
        [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
        None,
        id="Multiple actions, no overlapping scope",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All()),
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([1])),
        ],
        [TimeSeriesAcl.Action.Read],
        [AllScope()],
        id="AllScope present with other scopes",
    )
    yield pytest.param(
        [
            TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All()),
        ],
        [AssetsAcl.Action.Read],
        None,
        id="Unknown action class",
    )
    yield pytest.param(
        [
            DataModelInstancesAcl(
                [DataModelInstancesAcl.Action.Read], DataModelInstancesAcl.Scope.SpaceID(["space1", "space2", "space3"])
            ),
            DataModelInstancesAcl(
                [DataModelInstancesAcl.Action.Write],
                DataModelInstancesAcl.Scope.SpaceID(["space2", "space4", "space5"]),
            ),
            DataModelInstancesAcl(
                [DataModelInstancesAcl.Action.Write_Properties],
                DataModelInstancesAcl.Scope.SpaceID(["space2", "space3", "space6"]),
            ),
        ],
        [
            DataModelInstancesAcl.Action.Read,
            DataModelInstancesAcl.Action.Write,
            DataModelInstancesAcl.Action.Write_Properties,
        ],
        [SpaceIDScope(space_ids=["space2"])],
        id="DataModelInstancesAcl with multiple actions and SpaceID scope intersection",
    )
    yield pytest.param(
        [
            AssetsAcl([AssetsAcl.Action.Read], AssetsAcl.Scope.DataSet([1])),
            AssetsAcl([AssetsAcl.Action.Write], AssetsAcl.Scope.DataSet([2])),
        ],
        [AssetsAcl.Action.Read, AssetsAcl.Action.Write],
        None,
        id="AssetsAcl with multiple actions, no overlapping scope",
    )
    yield pytest.param(
        [TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.DataSet([123]))],
        [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
        None,
        id="One of multiple actions is missing capability",
    )


class TestTokenAPI:
    @pytest.mark.parametrize(
        "existing_capabilities,actions,expected_scope",
        list(get_scope_test_cases()),
    )
    def test_get_scope(
        self,
        existing_capabilities: list[Capability],
        actions: list[Capability.Action],
        expected_scope: list[Capability.Scope] | None,
    ):
        with monkeypatch_toolkit_client() as client:
            client.iam.token.inspect.return_value = TokenInspection(
                subject="test_subject",
                projects=[ProjectSpec("https://example.com", groups=[123])],
                capabilities=ProjectCapabilityList(
                    [ProjectCapability(cap, ProjectsScope(["my_project"])) for cap in existing_capabilities]
                ),
            )

            actual_scope = client.token.get_scope(actions)
            assert actual_scope == expected_scope

    @pytest.mark.parametrize(
        "actions, error_message",
        [
            pytest.param([], "No actions provided to get_scope", id="Empty actions list"),
            pytest.param(
                [TimeSeriesAcl.Action.Read, AssetsAcl.Action.Write],
                "Actions belong to multiple ACL classes: AssetsAcl and TimeSeriesAcl.",
                id="Multiple actions with different ACL classes",
            ),
            pytest.param([object], "Unknown action class: type", id="Unknown action class"),
        ],
    )
    def test_get_scope_invalid_inputs(self, actions: list[Capability.Action], error_message: str) -> None:
        with monkeypatch_toolkit_client() as client:
            with pytest.raises(ValueError) as exc_info:
                client.token.get_scope(actions)

        assert str(exc_info.value) == error_message



from collections.abc import Sequence

from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.api_classes import PagedResponse
from cognite_toolkit._cdf_tk.client.data_classes.streams import StreamRequest, StreamResponse
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ItemsRequest, ParamRequest


class StreamsAPI:
    ENDPOINT = "/streams"

    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self._http_client = http_client
        self._console = console
        self._config = http_client.config

    def create(self, items: Sequence[StreamRequest]) -> list[StreamResponse]:
        """Create one or more streams.

        Args:
            items: Sequence of StreamRequest items to create.

        Returns:
            List of created StreamResponse items.
        """
        responses = self._http_client.request_with_retries(
            ItemsRequest(
                endpoint_url=self._config.create_api_url(self.ENDPOINT),
                method="POST",
                items=list(items),
            )
        )
        responses.raise_for_status()
        return PagedResponse[StreamResponse].model_validate(responses.get_first_body()).items

    def delete(self, external_id: str) -> None:
        """Delete stream using its external ID.

        Args:
            external_id: External ID of the stream to delete.
        """
        responses = self._http_client.request_with_retries(
            ParamRequest(
                endpoint_url=self._config.create_api_url(f"{self.ENDPOINT}/{external_id}"),
                method="DELETE",
            )
        )
        responses.raise_for_status()

    def list(self) -> list[StreamResponse]:
        """List streams.

        Returns:
            StreamResponseList containing the listed streams.
        """
        responses = self._http_client.request_with_retries(
            ParamRequest(
                endpoint_url=self._config.create_api_url(self.ENDPOINT),
                method="GET",
            )
        )
        responses.raise_for_status()
        return PagedResponse[StreamResponse].model_validate(responses.get_first_body()).items

    def retrieve(self, external_id: str, include_statistics: bool = True) -> StreamResponse:
        """Retrieve a stream by its external ID.

        Args:
            external_id: External ID of the stream to retrieve.
            include_statistics: Whether to include usage statistics in the response.
        Returns:
            StreamResponse item.
        """
        responses = self._http_client.request_with_retries(
            ParamRequest(
                endpoint_url=self._config.create_api_url(f"{self.ENDPOINT}/{external_id}"),
                method="GET",
                parameters={"includeStatistics": include_statistics},
            )
        )
        responses.raise_for_status()
        response_body = responses.get_first_body()
        return StreamResponse.model_validate(response_body)



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.transformation_notification import TransformationNotificationYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestTransformationNotificationYAML:
    @pytest.mark.parametrize("data", list(find_resources("Notification", "transformations")))
    def test_load_valid_transformation_notification(self, data: dict[str, object]) -> None:
        loaded = TransformationNotificationYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data, expected_errors",
        [
            pytest.param(
                {"destination": "example@email.com"},
                {"Missing required field: 'transformationExternalId'"},
                id="missing_required_field",
            ),
            pytest.param(
                {"transformationExternalId": 123, "destination": "example@emai.com"},
                {
                    "In field transformationExternalId input should be a valid string. Got 123 of "
                    "type int. Hint: Use double quotes to force string."
                },
                id="wrong_type_field",
            ),
            pytest.param(
                {
                    "destination": "example@email.com",
                    "transformationId": 123,
                    "transformationExternalId": "ext-123",
                },
                {"Unused field: 'transformationId'"},
                id="Specifying transformationId",
            ),
        ],
    )
    def test_load_invalid_transformation_notification(self, data: dict[str, object], expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, TransformationNotificationYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    Frame,
    FrameList,
    FrameWrite,
    _FrameUpdate,
)

from .utlis import tmp_disable_gzip


class FramesAPI(APIClient):
    _RESOURCE_PATH = "/robotics/frames"

    @overload
    def __call__(self) -> Iterator[Frame]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[FrameList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[Frame] | Iterator[FrameList]:
        """Iterate over robot frames.

        Args:
            chunk_size: The number of robot frames to return in each chunk. None will return all robot frames.

        Yields:
            Frame or FrameList

        """
        return self._list_generator(method="GET", resource_cls=Frame, list_cls=FrameList, chunk_size=chunk_size)

    def __iter__(self) -> Iterator[Frame]:
        return self.__call__()

    @overload
    def create(self, frame: FrameWrite) -> Frame: ...

    @overload
    def create(self, frame: Sequence[FrameWrite]) -> FrameList: ...

    def create(self, frame: FrameWrite | Sequence[FrameWrite]) -> Frame | FrameList:
        """Create a new robot frame.

        Args:
            frame: FrameWrite or list of FrameWrite.

        Returns:
            Frame object.

        """
        with tmp_disable_gzip():
            return self._create_multiple(
                list_cls=FrameList,
                resource_cls=Frame,
                items=frame,
                input_resource_cls=FrameWrite,
            )

    @overload
    def retrieve(self, external_id: str) -> Frame | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> FrameList: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> Frame | None | FrameList:
        """Retrieve a robot frame.

        Args:
            external_id: External id of the robot frame.

        Returns:
            Frame object.

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            return self._retrieve_multiple(
                identifiers=identifiers,
                resource_cls=Frame,
                list_cls=FrameList,
            )

    @overload
    def update(self, frame: FrameWrite) -> Frame: ...

    @overload
    def update(self, frame: Sequence[FrameWrite]) -> FrameList: ...

    def update(self, frame: FrameWrite | Sequence[FrameWrite]) -> Frame | FrameList:
        """Update a robot frame.

        Args:
            frame: FrameWrite or list of FrameWrite.

        Returns:
            Frame object.

        """
        with tmp_disable_gzip():
            return self._update_multiple(
                items=frame,
                resource_cls=Frame,
                list_cls=FrameList,
                update_cls=_FrameUpdate,
            )

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete a robot frame.

        Args:
            external_id: External id of the robot frame.

        Returns:
            None

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            self._delete_multiple(identifiers=identifiers, wrap_ids=True)

    def list(self) -> FrameList:
        """List robot frames.

        Returns:
            FrameList

        """
        with tmp_disable_gzip():
            return self._list(method="GET", resource_cls=Frame, list_cls=FrameList)



import difflib
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, ClassVar, cast

from cognite_toolkit._cdf_tk.constants import INDEX_PATTERN
from cognite_toolkit._cdf_tk.cruds import (
    CRUDS_BY_FOLDER_NAME,
    GroupCRUD,
    RawDatabaseCRUD,
    RawTableCRUD,
    ResourceCRUD,
)
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    BuiltResourceList,
    ModuleLocation,
)
from cognite_toolkit._cdf_tk.exceptions import (
    AmbiguousResourceFileError,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    ToolkitNotSupportedWarning,
    ToolkitWarning,
    WarningList,
)
from cognite_toolkit._cdf_tk.tk_warnings.fileread import (
    UnknownResourceTypeWarning,
)
from cognite_toolkit._cdf_tk.utils import (
    humanize_collection,
    safe_read,
)


class Builder(ABC):
    _resource_folder: ClassVar[str | None] = None

    def __init__(
        self,
        build_dir: Path | None,
        resource_folder: str | None = None,
        warn: Callable[[ToolkitWarning], None] | None = None,
    ):
        self._build_dir = build_dir
        self.warn = warn
        self.resource_counter = 0
        if self._resource_folder is not None:
            self.resource_folder = self._resource_folder
        elif resource_folder is not None:
            self.resource_folder = resource_folder
        else:
            raise ValueError("Either _resource_folder or resource_folder must be set.")

    @property
    def build_dir(self) -> Path:
        if self._build_dir is None:
            raise ValueError("build_dir must be set for this operation.")
        return self._build_dir

    @abstractmethod
    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | Sequence[ToolkitWarning]]:
        raise NotImplementedError()

    def load_extra_field(self, extra: str) -> tuple[str, Any]:
        """Overload in subclass to load extra fields from a file."""
        raise NotImplementedError(
            f"Extra field {extra!r} by {type(self).__name__} - {self.resource_folder} is not supported."
        )

    def validate_directory(
        self, built_resources: BuiltResourceList, module: ModuleLocation
    ) -> WarningList[ToolkitWarning]:
        """This can be overridden to add additional validation for the built resources."""
        return WarningList[ToolkitWarning]()

    # Helper methods
    def _create_destination_path(self, source_path: Path, kind: str) -> Path:
        """Creates the filepath in the build directory for the given source path.

        Note that this is a complex operation as the modules in the source are nested while the build directory is flat.
        This means that we lose information and risk having duplicate filenames. To avoid this, we prefix the filename
        with a number to ensure uniqueness.
        """
        filestem = source_path.stem
        # Get rid of the local index
        filestem = INDEX_PATTERN.sub("", filestem)

        # Increment to ensure we do not get duplicate filenames when we flatten the file
        # structure from the module to the build directory.
        self.resource_counter += 1

        filename = f"{self.resource_counter}.{filestem}"
        if not filename.casefold().endswith(kind.casefold()):
            filename = f"{filename}.{kind}"
        filename = f"{filename}{source_path.suffix}"
        destination_path = self.build_dir / self.resource_folder / filename
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        return destination_path

    def _get_loader(self, source_path: Path) -> tuple[None, ToolkitWarning] | tuple[type[ResourceCRUD], None]:
        return get_loader(source_path, self.resource_folder)


def get_loader(
    source_path: Path,
    resource_folder: str,
    force_pattern: bool = False,
) -> tuple[None, ToolkitWarning] | tuple[type[ResourceCRUD], None]:
    folder_loaders = CRUDS_BY_FOLDER_NAME.get(resource_folder, [])
    if not folder_loaders:
        return None, ToolkitNotSupportedWarning(
            f"resource of type {resource_folder!r} in {source_path.name}.",
            details=f"Available resources are: {', '.join(CRUDS_BY_FOLDER_NAME.keys())}",
        )

    loaders = [
        loader for loader in folder_loaders if loader.is_supported_file(source_path, force_pattern=force_pattern)
    ]
    if len(loaders) == 0:
        suggestion: str | None = None
        if "." in source_path.stem:
            core, kind = source_path.stem.rsplit(".", 1)
            match = difflib.get_close_matches(kind, [loader.kind for loader in folder_loaders])
            if match:
                suggested_name = f"{core}.{match[0]}{source_path.suffix}"
                suggestion = f"Did you mean to call the file {suggested_name!r}?"
        else:
            kinds = [loader.kind for loader in folder_loaders]
            if len(kinds) == 1:
                suggestion = f"Did you mean to call the file '{source_path.stem}.{kinds[0]}{source_path.suffix}'?"
            else:
                suggestion = (
                    f"All files in the {resource_folder!r} folder must have a file extension that matches "
                    f"the resource type. Supported types are: {humanize_collection(kinds)}."
                )
        return None, UnknownResourceTypeWarning(source_path, suggestion)
    elif len(loaders) > 1 and all(loader.folder_name == "raw" for loader in loaders):
        # Raw files can be ambiguous, so we need to check the content.
        # If there is a tableName field, it is a table, otherwise it is a database.
        if any(
            line.strip().startswith("tableName:") or line.strip().startswith("- tableName:")
            for line in safe_read(source_path).splitlines()
        ):
            return RawTableCRUD, None
        else:
            return RawDatabaseCRUD, None
    elif len(loaders) > 1 and all(issubclass(loader, GroupCRUD) for loader in loaders):
        # There are two group loaders, one for resource scoped and one for all scoped.
        return GroupCRUD, None
    elif len(loaders) > 1:
        names = humanize_collection(
            [f"'{source_path.stem}.{loader.kind}{source_path.suffix}'" for loader in loaders], bind_word="or"
        )
        raise AmbiguousResourceFileError(
            f"Ambiguous resource file {source_path.name} in {resource_folder} folder. "
            f"Unclear whether it is {humanize_collection([loader.kind for loader in loaders], bind_word='or')}."
            f"\nPlease name the file {names}."
        )

    return cast(type[ResourceCRUD], loaders[0]), None


class DefaultBuilder(Builder):
    """This is used to build resources that do not have a specific builder."""

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | list[ToolkitWarning]]:
        for source_file in source_files:
            if source_file.loaded is None:
                # Not a YAML file
                continue
            loader, warning = self._get_loader(source_file.source.path)
            if loader is None:
                if warning is not None:
                    yield [warning]
                continue
            destination_path = self._create_destination_path(source_file.source.path, loader.kind)

            destination = BuildDestinationFile(
                path=destination_path,
                loaded=source_file.loaded,
                loader=loader,
                source=source_file.source,
                extra_sources=None,
            )
            yield destination



import sys
from collections.abc import Sequence
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes import GeoLocation, Label
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.files import FileMetadata, FileMetadataList

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ExtendedFileMetadata(FileMetadata):
    """Extended FileMetadata with pending instance ID support.
    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        instance_id (NodeId | None): The Instance ID for the file. (Only applicable for files created in DMS)
        name (str | None): Name of the file.
        source (str | None): The source of the file.
        mime_type (str | None): File type. E.g., text/plain, application/pdf, ...
        metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value. Limits: Maximum length of key is 32 bytes, value 512 bytes, up to 16 key-value pairs.
        directory (str | None): Directory associated with the file. It must be an absolute, unix-style path.
        asset_ids (Sequence[int] | None): No description.
        data_set_id (int | None): The dataSet ID for the item.
        labels (Sequence[Label] | None): A list of the labels associated with this resource item.
        geo_location (GeoLocation | None): The geographic metadata of the file.
        source_created_time (int | None): The timestamp for when the file was originally created in the source system.
        source_modified_time (int | None): The timestamp for when the file was last modified in the source system.
        security_categories (Sequence[int] | None): The security category IDs required to access this file.
        id (int | None): A server-generated ID for the object.
        uploaded (bool | None): Whether the actual file is uploaded. This field is returned only by the API, it has no effect in a post body.
        uploaded_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        pending_instance_id( NodeId | None): The pending instance ID for the file, used in upgrading files from asset centric to data modeling.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        external_id: str | None = None,
        instance_id: NodeId | None = None,
        name: str | None = None,
        source: str | None = None,
        mime_type: str | None = None,
        metadata: dict[str, str] | None = None,
        directory: str | None = None,
        asset_ids: Sequence[int] | None = None,
        data_set_id: int | None = None,
        labels: Sequence[Label] | None = None,
        geo_location: GeoLocation | None = None,
        source_created_time: int | None = None,
        source_modified_time: int | None = None,
        security_categories: Sequence[int] | None = None,
        id: int | None = None,
        uploaded: bool | None = None,
        uploaded_time: int | None = None,
        created_time: int | None = None,
        last_updated_time: int | None = None,
        pending_instance_id: NodeId | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            instance_id=instance_id,
            name=name,
            directory=directory,
            source=source,
            mime_type=mime_type,
            metadata=metadata,
            asset_ids=asset_ids,
            data_set_id=data_set_id,
            labels=labels,
            geo_location=geo_location,
            source_created_time=source_created_time,
            source_modified_time=source_modified_time,
            security_categories=security_categories,
            id=id,
            uploaded=uploaded,
            uploaded_time=uploaded_time,
            created_time=created_time,
            last_updated_time=last_updated_time,
            cognite_client=cognite_client,
        )
        self.pending_instance_id = pending_instance_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client)
        if isinstance(instance.pending_instance_id, dict):
            instance.pending_instance_id = NodeId.load(instance.pending_instance_id)
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the object to a dictionary"""
        output = super().dump(camel_case=camel_case)
        if self.pending_instance_id is not None:
            output["pendingInstanceId" if camel_case else "pending_instance_id"] = self.pending_instance_id.dump(
                camel_case=camel_case, include_instance_type=False
            )
        return output


class ExtendedFileMetadataList(FileMetadataList):
    _RESOURCE = ExtendedFileMetadata  # type: ignore[assignment]



from pydantic import Field, SecretStr

from .base import BaseModelResource


class AuthenticationClientIdSecret(BaseModelResource):
    client_id: str = Field(description="Client Id.")
    client_secret: SecretStr = Field(description="Client Secret.")


class OIDCCredential(AuthenticationClientIdSecret):
    scopes: str | list[str] | None = Field(default=None, description="Scopes for the authentication.")
    token_uri: str = Field(description="OAuth token url.")
    cdf_project_name: str = Field(description="CDF project name.")
    audience: str | None = Field(default=None, description="Audience for the authentication.")



import json
from typing import Any, ClassVar

from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.utils._identifier import InstanceId
from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel, Field, field_serializer

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


class MigrationObject(BaseModel, alias_generator=to_camel_case, extra="ignore", populate_by_name=True): ...


class MigrationIssue(MigrationObject):
    """Represents an issue encountered during migration."""

    type: ClassVar[str]

    def dump(self) -> dict[str, JsonVal]:
        # Dump json to ensure it is serializable
        dumped = json.loads(self.model_dump_json(by_alias=True))
        dumped["type"] = self.type
        return dumped

    @property
    def has_issues(self) -> bool:
        """Check if there are any issues recorded in this MigrationIssue."""
        return True


class ReadIssue(MigrationIssue):
    """Represents a read issue encountered during migration."""

    type: ClassVar[str] = "read"


class ReadFileIssue(ReadIssue):
    """Represents a read issue encountered during migration of a file.

    Attributes:
        row_no (int): The row number in the CSV file where the issue occurred.
        error (str | None): An optional error message providing additional details about the read issue.
    """

    type: ClassVar[str] = "fileRead"

    row_no: int
    error: str | None = None


class ReadAPIIssue(ReadIssue):
    """Represents a read issue encountered during migration from the API.

    Attributes:
        asset_centric_id (AssetCentricId): The identifier of the asset-centric resource that could not be read.
        error (str | None): An optional error message providing additional details about the read issue.
    """

    type: ClassVar[str] = "apiRead"
    asset_centric_id: AssetCentricId
    error: str | None = None

    @field_serializer("asset_centric_id")
    def serialize_asset_centric_id(self, asset_centric_id: AssetCentricId) -> dict[str, Any]:
        return {
            "resourceType": asset_centric_id.resource_type,
            "id": asset_centric_id.id_,
        }


class FailedConversion(MigrationObject):
    """Represents a property that failed to convert during migration.

    Attributes:
        value (str | int | float | bool | None): The value that failed to convert
        error (str): The error message explaining why the conversion failed.

    """

    property_id: str
    value: str | int | float | bool | None | list | dict
    error: str


class InvalidPropertyDataType(MigrationObject):
    """Represents a property with an invalid type during migration.

    Attributes:
        property_id (str): The identifier of the property in asset-centric.

        expected_type (str): The expected type of the property.
    """

    property_id: str
    expected_type: str


class ConversionIssue(MigrationIssue):
    """Represents a conversion issue encountered during migration.

    Attributes:
        asset_centric_id (AssetCentricId): The identifier of the asset-centric resource.
        instance_id (InstanceId): The NodeId of the data model instance.
        missing_asset_centric_properties (list[str]): List of source properties that are missing.
        missing_instance_properties (list[str]): List of target properties that are missing.
        invalid_instance_property_types (list[InvalidPropertyDataType]): List of properties with invalid types.
        failed_conversions (list[FailedConversion]): List of properties that failed to convert with reasons.
    """

    type: ClassVar[str] = "conversion"
    asset_centric_id: AssetCentricId
    instance_id: InstanceId
    missing_asset_centric_properties: list[str] = Field(default_factory=list)
    missing_instance_properties: list[str] = Field(default_factory=list)
    invalid_instance_property_types: list[InvalidPropertyDataType] = Field(default_factory=list)
    failed_conversions: list[FailedConversion] = Field(default_factory=list)
    ignored_asset_centric_properties: list[str] = Field(default_factory=list)
    missing_instance_space: str | None = None

    @property
    def has_issues(self) -> bool:
        """Check if there are any issues recorded in this ConversionIssue."""
        return bool(
            self.missing_asset_centric_properties
            or self.missing_instance_properties
            or self.invalid_instance_property_types
            or self.failed_conversions
            or self.missing_instance_space
        )

    @field_serializer("instance_id")
    def serialize_instance_id(self, instance_id: NodeId) -> dict[str, str]:
        return instance_id.dump(include_instance_type=True)

    @field_serializer("asset_centric_id")
    def serialize_asset_centric_id(self, asset_centric_id: AssetCentricId) -> dict[str, Any]:
        return {
            "resourceType": asset_centric_id.resource_type,
            "id": asset_centric_id.id_,
        }


class WriteIssue(MigrationIssue):
    """Represents a write issue encountered during migration.

    Attributes:
        instance_id (InstanceId): The InstanceId of the data model instance that could not be written.
        status_code (int): The HTTP status code returned during the write operation.
        message (str | None): An optional message providing additional details about the write issue.
    """

    type: ClassVar[str] = "write"
    instance_id: InstanceId
    status_code: int
    message: str | None = None

    @field_serializer("instance_id")
    def serialize_instance_id(self, instance_id: NodeId) -> dict[str, str]:
        return instance_id.dump(include_instance_type=True)



from urllib.parse import urljoin

from cognite.client import CogniteClient
from cognite.client._api_client import APIClient
from cognite.client.config import ClientConfig

from cognite_toolkit._cdf_tk.client.data_classes.search_config import SearchConfig, SearchConfigList, SearchConfigWrite


class SearchConfigurationsAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)

    def _get_base_url_with_base_path(self) -> str:
        """
        This method is overridden to provide the correct base path for the Search Configurations API.
        This method in base class APIClient appends /api/{api_version}/ to the base URL,
        but for Search Configurations API, we need a different path structure.
        """
        base_path = ""
        if self._api_version:
            base_path = f"/apps/{self._api_version}/projects/{self._config.project}/storage/config/apps/search/views"
        return urljoin(self._config.base_url, base_path)

    def upsert(self, configuration_update: SearchConfigWrite) -> SearchConfig:
        """Update/Create a Configuration.

        Args:
            configuration_update: The content of the configuration to update

        Returns:
            SearchConfig

        """
        res = self._post(
            url_path="/upsert",
            json=configuration_update.dump(),
        )
        return SearchConfig._load(res.json(), cognite_client=self._cognite_client)

    def list(self) -> SearchConfigList:
        """List all Configuration.

        Returns:
            SearchConfigList

        """
        res = self._post(url_path="/list")
        return SearchConfigList._load(res.json()["items"], cognite_client=self._cognite_client)
